"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of functions around exporting from Revit to nwc file format.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from Autodesk.Revit.DB import ElementId, IFCExportOptions, Transaction, ViewType

from duHast.Revit.Common import transaction as rTran
from duHast.Revit.Common.revit_version import get_revit_version_number
from duHast.Utilities.Objects import result as res
from duHast.Revit.Views import views as rView
from duHast.Revit.Exports.export import build_export_file_name_from_view
from duHast.Revit.Exports.Utility.ifc_export_coordinates import IFCCoords
from duHast.Revit.Exports.Utility.ifc_export_space_boundaries import IFCSpaceBoundaries

from duHast.Revit.Exports.Utility.export_ifc_config_2019 import (
    ifc_get_third_party_export_config_by_model_2019,
    ifc_get_third_party_export_config_by_view_2019,
)
from duHast.Revit.Exports.Utility.export_ifc_config_2020 import (
    ifc_get_third_party_export_config_by_model_2020,
    ifc_get_third_party_export_config_by_view_2020,
)
from duHast.Revit.Exports.Utility.export_ifc_config_2021 import (
    ifc_get_third_party_export_config_by_model_2021,
    ifc_get_third_party_export_config_by_view_2021,
)
from duHast.Revit.Exports.Utility.export_ifc_config_2022 import (
    ifc_get_third_party_export_config_by_model_2022,
    ifc_get_third_party_export_config_by_view_2022,
)
from duHast.Revit.Exports.Utility.export_ifc_config_2023 import (
    ifc_get_third_party_export_config_by_model_2023,
    ifc_get_third_party_export_config_by_view_2023,
)

from duHast.Revit.Exports.Utility.export_ifc_config_2024 import (
    ifc_get_third_party_export_config_by_model_2024,
    ifc_get_third_party_export_config_by_view_2024,
)


def ifc_get_export_config_by_view(
    ifc_version, ifc_space_bounds=IFCSpaceBoundaries.no_boundaries
):
    """
    Returns an IFC export configuration for the built in IFC exporter.
    :param ifc_version: The ifc version used for the export.
    :type ifc_version: Autodesk.Revit.DB.IFCVersion
    :param ifc_space_bounds: IFC space boundary setting, defaults to IFCSpaceBoundaries.noBoundaries
    :type ifc_space_bounds: SampleCodeBatchProcessor.RevitExport.IFCSpaceBoundaries, optional
    :return: an IFC export option
    :rtype: Autodesk.Revit.DB.IFCExportOptions
    """

    ex_ifc = IFCExportOptions()
    ex_ifc.ExportBaseQuantities = True
    ex_ifc.FileVersion = ifc_version
    ex_ifc.SpaceBoundaryLevel = ifc_space_bounds
    ex_ifc.WallAndColumnSplitting = True
    return ex_ifc


def export_to_ifc(doc, ifc_export_option, directory_path, file_name):
    """
    Exports to IFC either the entire model or a view only using 3rd party exporter.
    What gets exported is defined in the ifc_export_option.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param ifc_export_option: The settings for the IFC export.
    :type ifc_export_option: Autodesk.Revit.DB IFCExportOptions
    :param directory_path: The directory path to where the export is being saved.
    :type directory_path: str
    :param file_name: The file name under which the export is being saved.
    :type file_name: str
    :return:
        Result class instance.
        - Export status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain the fully qualified file path of the exported file.
        On exception:
        - result.status (bool) will be False.
        - result.message will contain the exception message.
    :rtype: :class:`.Result`
    """

    # ifc export needs to run in a transaction
    return_value = res.Result()

    def action():
        action_return_value = res.Result()
        try:
            # export to IFC
            doc.Export(directory_path, file_name, ifc_export_option)
            action_return_value.update_sep(
                True, "Exported: {}".format(directory_path + "\\" + str(file_name))
            )
            # needs to be a list in a list to stay together when combined with previous results in the update status result code
            action_return_value.result = [[directory_path, file_name]]
        except Exception as e:
            action_return_value.update_sep(
                False,
                "Script Exception: Failed to export to IFC with exception: {}".format(
                    e
                ),
            )
        return action_return_value

    transaction = Transaction(doc, "Export to IFC")
    return_value = rTran.in_transaction(transaction, action)
    return return_value


def export_3d_views_to_ifc_default(doc, view_filter, ifc_export_option, directory_path):
    """
    Exports 3D views matching a filter (view starts with) to IFC using the default built-in exporter.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param view_filter: String the view name is to start with if it is to be exported. (Both view name and string are set to lowercase for comparison)
    :type view_filter: str
    :param ifc_export_option: The IFC export option.
    :type ifc_export_option: Autodesk.Revit.DB.IFCExportOptions
    :param directory_path: The directory path to where the export is being saved.
    :type directory_path: str

    :return: Result class instance.
        - `result.status` (bool): False if an exception occurred, otherwise True (export status).
        - `result.message` (str): Contains the fully qualified file path of the exported file.

    On exception:
        - `result.status` (bool): False.
        - `result.message` (str): Contains the exception message.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    views_to_export = []
    # get all 3D views in model and filter out views to be exported
    views = rView.get_views_of_type(doc, ViewType.ThreeD)
    for v in views:
        if v.Name.lower().startswith(view_filter.lower()):
            views_to_export.append(v)
    # export those views one by one
    if len(views_to_export) > 0:
        for export_view in views_to_export:
            return_value_by_view = res.Result()
            ifc_export_option.FilterViewId = export_view.Id
            file_name = build_export_file_name_from_view(
                export_view.Name, view_filter, ".ifc"
            )
            return_value_by_view = export_to_ifc(
                doc, ifc_export_option, directory_path, file_name
            )
            return_value.update(return_value_by_view)
    else:
        return_value.update_sep(
            True, "No 3D views found matching filter...nothing was exported"
        )
    return return_value


def setup_ifc_export_option(
    export_config,
    view_id=ElementId.InvalidElementId,
    coord_option=IFCCoords.shared_coordinates,
    convert_view_id_to_integer=True,
):
    """
    Function assigning a view Id to export ifc config if it is exporting by view.
    By model export will assign Autodesk.Revit.DB.ElementId.InvalidElementId instead.
    :param export_config: The ifc export configuration used.
    :type export_config: BIM.IFC.Export.UI.IFCExportConfiguration
    :param view_id: The id of the view to be exported, defaults to ElementId.InvalidElementId
    :type view_id: Autodesk.Revit.DB.ElementId, optional
    :param coord_option: Describes the coordinate options used for the export, defaults to IFCCoords.SharedCoordinates
    :type coord_option: SampleCodeBatchProcessor.RevitExport.IFCCoords, optional
    :return: An updated ifc export option instance.
    :rtype: Autodesk.Revit.DB.IFCExportOptions
    """

    if export_config.UseActiveViewGeometry == True:
        if convert_view_id_to_integer:
            export_config.ActiveViewId = view_id.IntegerValue
        else:
            export_config.ActiveViewId = view_id
    else:
        if convert_view_id_to_integer:
            export_config.ActiveViewId = -1
        else:
            export_config.ActiveViewId = ElementId.InvalidElementId

    # set up the ifc export options object
    ex_ifc = IFCExportOptions()

    # do I need to convert the view id to int here as well if required?
    export_config.UpdateOptions(ex_ifc, view_id)

    # set the coordinate system to use
    ex_ifc.AddOption("SitePlacement", coord_option)

    return ex_ifc


def export_3d_views_to_ifc(
    doc,
    view_filter,
    ifc_export_config,
    directory_path,
    ifc_coordinates_system=IFCCoords.shared_coordinates,
    do_something_with_view_name=None,
):
    """
    Exports 3D views matching a filter (view starts with) to IFC using a 3rd party exporter.

    By default, the file name of the export will be the same as the name of the view being exported.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param view_filter: String the view name is to start with if it is to be exported. (Both view name and string are set to lowercase for comparison)
    :type view_filter: str
    :param ifc_export_config: The IFC export configuration.
    :type ifc_export_config: BIM.IFC.Export.UI.IFCExportConfiguration
    :param directory_path: The directory path to where the export is being saved.
    :type directory_path: str
    :param ifc_coordinates_system: Describes the coordinate options used for the export, defaults to IFCCoords.SharedCoordinates.
    :type ifc_coordinates_system: SampleCodeBatchProcessor.RevitExport.IFCCoords, optional
    :param do_something_with_view_name: A function that takes the view name as an argument and processes it. The modified view name is then used as the actual file name, defaults to None, which uses the view name unchanged as the export file name.
    :type do_something_with_view_name: function, optional

    :return: Result class instance.
        - `result.status` (bool): False if an exception occurred, otherwise True (export status).
        - `result.message` (str): Contains the fully qualified file path of the exported file.

    On exception:
        - `result.status` (bool): False.
        - `result.message` (str): Contains the exception message.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    # get the revit version:
    convert_view_id_to_integer = True
    # revit 2024 expects an Element Id for the view id not the integer value of the id
    revit_version = get_revit_version_number(doc=doc)
    if revit_version == 2024:
        convert_view_id_to_integer = False

    views_to_export = []
    # get all 3D views in model and filter out views to be exported
    views = rView.get_views_of_type(doc, ViewType.ThreeD)
    for v in views:
        if v.Name.lower().startswith(view_filter.lower()):
            views_to_export.append(v)
    # export those views one by one
    if len(views_to_export) > 0:
        for export_view in views_to_export:
            return_value_by_view = res.Result()
            updated_export_option = setup_ifc_export_option(
                export_config=ifc_export_config,
                view_id=export_view.Id,
                coord_option=ifc_coordinates_system,
                convert_view_id_to_integer=convert_view_id_to_integer,
            )
            file_name = (
                build_export_file_name_from_view(export_view.Name, view_filter, ".ifc")
                if do_something_with_view_name == None
                else do_something_with_view_name(export_view.Name)
            )
            return_value_by_view = export_to_ifc(
                doc, updated_export_option, directory_path, file_name
            )
            return_value.update(return_value_by_view)
    else:
        return_value.update_sep(
            True, "No 3D views found matching filter...nothing was exported"
        )
    return return_value


def export_model_to_ifc(
    doc,
    ifc_export_config,
    directory_path,
    file_name,
    coord_option=IFCCoords.shared_coordinates,
):
    """
    Function exporting the entire model to IFC using 3rd party exporter.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param ifc_export_config: The IFC export configuration.
    :type ifc_export_config: BIM.IFC.Export.UI.IFCExportConfiguration
    :param directory_path: The directory path to where the export is being saved.
    :type directory_path: str
    :param file_name: The file name under which the export is being saved.
    :type file_name: str
    :param coord_option: Describes the coordinate options used for the export, defaults to IFCCoords.SharedCoordinates
    :type coord_option: SampleCodeBatchProcessor.RevitExport.IFCCoords, optional

    :return:
        Result class instance.
        - Export status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain the fully qualified file path of the exported file.
        On exception:
        - result.status (bool) will be False.
        - result.message will contain the exception message.
    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    # get the revit version:
    convert_view_id_to_integer = True
    # revit 2024 expects an Element Id for the view id not the integer value
    revit_version = get_revit_version_number(doc=doc)
    if revit_version == 2024:
        convert_view_id_to_integer = False

    # need to create an export option from the export config
    ifc_export_option = setup_ifc_export_option(
        export_config=ifc_export_config,
        view_id=ElementId.InvalidElementId,
        coord_option=coord_option,
        convert_view_id_to_integer=convert_view_id_to_integer,
    )

    return_value_by_model = export_to_ifc(
        doc, ifc_export_option, directory_path, file_name
    )
    return_value.update(return_value_by_model)
    return return_value


def ifc_get_third_party_export_config_by_model(doc, ifc_version, ifc_settings=None):
    """
    Returns the 3rd party ifc config for export by model depending on revit version.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param ifc_version: The ifc version used for the export.
    :type ifc_version: Autodesk.Revit.DB.IFCVersion
    :raises ValueError: Raises an exception if the revit version in use is not supported by this script.
    :return: An ifc export config instance based on the Revit version.
    :rtype: BIM.IFC.Export.UI.IFCExportConfiguration
    """

    # get the revit version:
    revit_version = get_revit_version_number(doc=doc)
    ifc_config = None
    if revit_version == 2019:
        ifc_config = ifc_get_third_party_export_config_by_model_2019(
            ifc_version, ifc_settings
        )
    elif revit_version == 2020:
        ifc_config = ifc_get_third_party_export_config_by_model_2020(
            ifc_version, ifc_settings
        )
    elif revit_version == 2021:
        ifc_config = ifc_get_third_party_export_config_by_model_2021(
            ifc_version, ifc_settings
        )
    elif revit_version == 2022:
        ifc_config = ifc_get_third_party_export_config_by_model_2022(
            ifc_version, ifc_settings
        )
    elif revit_version == 2023:
        ifc_config = ifc_get_third_party_export_config_by_model_2023(
            ifc_version=ifc_version, ifc_settings=ifc_settings
        )
    elif revit_version == 2024:
        ifc_config = ifc_get_third_party_export_config_by_model_2024(
            ifc_version=ifc_version, ifc_settings=ifc_settings
        )
    else:
        # this is a non supported revit version!
        raise ValueError(
            "Revit version "
            + revit_version
            + " is currently not supported by IFC exporter!"
        )
    return ifc_config


def ifc_get_third_party_export_config_by_view(doc, ifc_version, ifc_settings=None):
    """
    Returns the 3rd party ifc config for export by view depending on revit version.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param ifc_version: The ifc version used for the export.
    :type ifc_version: Autodesk.Revit.DB.IFCVersion
    :raises ValueError: Raises an exception if the revit version in use is not supported by this script.
    :return: An ifc export config instance based on the Revit version.
    :rtype: BIM.IFC.Export.UI.IFCExportConfiguration
    """

    # get the revit version:
    revit_version = get_revit_version_number(doc=doc)
    ifc_config = None
    if revit_version == 2019:
        ifc_config = ifc_get_third_party_export_config_by_view_2019(
            ifc_version, ifc_settings
        )
    elif revit_version == 2020:
        ifc_config = ifc_get_third_party_export_config_by_view_2020(
            ifc_version, ifc_settings
        )
    elif revit_version == 2021:
        ifc_config = ifc_get_third_party_export_config_by_view_2021(
            ifc_version, ifc_settings
        )
    elif revit_version == 2022:
        ifc_config = ifc_get_third_party_export_config_by_view_2022(
            ifc_version, ifc_settings
        )
    elif revit_version == 2023:
        ifc_config = ifc_get_third_party_export_config_by_view_2023(
            ifc_version=ifc_version, ifc_settings=ifc_settings
        )
    elif revit_version == 2024:
        ifc_config = ifc_get_third_party_export_config_by_view_2024(
            ifc_version=ifc_version, ifc_settings=ifc_settings
        )
    else:
        # this is a non supported revit version!
        raise ValueError(
            "Revit version "
            + revit_version
            + " is currently not supported by IFC exporter!"
        )
    return ifc_config



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit plan view creation. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from duHast.Utilities.Objects.result import Result
from duHast.Revit.Common.transaction import in_transaction_with_failure_handling
from duHast.Revit.Common.parameter_set_utils import set_builtin_parameter_without_transaction_wrapper_by_name
from duHast.Revit.Common.Objects.FailureHandlingConfiguration import (
    FailureHandlingConfig,
)

from duHast.Revit.Views.views_crop import apply_view_crop_without_transaction_wrapper, apply_view_bounding_box_crop_without_transaction_wrapper

from Autodesk.Revit.DB import  BoundingBoxXYZ, BuiltInParameter, CurveLoop, Transaction, ViewPlan


def create_view_plan(
        doc, 
        level, 
        view_type, 
        view_name = None, 
        phasing=None, 
        view_crop = None, 
        view_crop_visible = True
    ):
    """
    Creates a new plan view in the Revit document.

    This function creates a new plan view in the specified Revit document, sets its properties, and handles any errors that occur during the process. The function supports setting the view name, phasing, and crop box properties.

    :param doc: The Revit document where the view will be created.
    :type doc: Autodesk.Revit.DB.Document
    :param level: The level at which the view will be created.
    :type level: Autodesk.Revit.DB.Level
    :param view_type: The type of the view to be created.
    :type view_type: Autodesk.Revit.DB.ViewFamilyType
    :param view_name: The name to assign to the new view, defaults to None.
    :type view_name: str, optional
    :param phasing: The phase to assign to the new view, defaults to None.
    :type phasing: Autodesk.Revit.DB.Element, optional
    :param view_crop: The crop box or loop to apply to the view, defaults to None.
    :type view_crop: Autodesk.Revit.DB.BoundingBoxXYZ or Autodesk.Revit.DB.CurveLoop, optional
    :param view_crop_visible: Flag indicating whether the crop box should be visible, defaults to True.
    :type view_crop_visible: bool, optional
    
    :return: A Result object containing the outcome of the view creation process.
    :rtype: Result
    """

    return_value = Result()

    try:
        # set up an action which creates the sheet and assigns properties
        def action():
            action_return_value = Result()
            try:
                view_plan = ViewPlan.Create(doc, view_type.Id, level.Id)

                # check if sheet was actually created?
                if not view_plan:
                    raise ValueError("Failed to create new plan view.")
                try:
                    # set view standard properties
                    if(view_name):
                        view_plan.Name = view_name
                except Exception as e:
                    # only pass on the message to caller
                    action_return_value.append_message("Failed to set view name to {} with exception: {}".format(view_name, e))
                
                # set the phasing
                result_phase = set_builtin_parameter_without_transaction_wrapper_by_name(
                    element=view_plan, 
                    parameter_definition=BuiltInParameter.VIEW_PHASE, 
                    parameter_value=phasing.Id
                )
                # only pass on the message to caller
                action_return_value.append_message(result_phase.message)
               
                # set the view crop ( can be a bounding box or a loop )
                if(isinstance(view_crop,  BoundingBoxXYZ)):
                    # set the crop shape
                    result_bbox = apply_view_bounding_box_crop_without_transaction_wrapper(view=view_plan, bbox=view_crop)
                    # only pass on the message to caller
                    return_value.append_message(result_bbox.message)
                elif (isinstance(view_crop, CurveLoop)):
                    result_loop =  apply_view_crop_without_transaction_wrapper(view=view_plan, crop_loop=view_crop)
                    # only pass on the message to caller
                    return_value.append_message(result_loop.message)
                
                # make sure the view is cropped if a crop method was supplied
                if view_crop:
                    view_plan.CropBoxActive = True
                
                # set the visibility of the view crop
                view_plan.CropBoxVisible = view_crop_visible
                
                action_return_value.append_message("Successfully created view {}".format(view_plan.Name))
                action_return_value.result.append(view_plan)
            except Exception as e:
                action_return_value.update_sep(False,"{}".format(e))
            return action_return_value
        
        # define failure handling for the transaction ( roll back on any warnings or errors )
        failure_handling_settings = FailureHandlingConfig(
            roll_back_on_warning=True,
            print_warnings=False,
            roll_back_on_error=True,
            print_errors=False,
        )

        if(view_name):
            view_name_transaction = view_name
        else:
            view_name_transaction = "default revit view name"
        # set up the transaction
        trans = Transaction(doc, "creating view plan {}".format(view_name_transaction))

        # execute the transaction with failure handling
        result_transaction = in_transaction_with_failure_handling(
            transaction=trans,
            action=action,
            failure_config=failure_handling_settings,
        )

        # update the return value with the result of the transaction
        return_value.update(result_transaction)
    except Exception as e:
        return_value.update_sep(False, "failed to create plan view with exception: {}".format(e))

    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit revisions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr

clr.AddReference("System.Core")
from System import Linq

clr.ImportExtensions(Linq)

# import System
from collections import namedtuple

from duHast.Utilities.Objects import result as res
from duHast.Revit.Common import (
    transaction as rTran,
    revit_version as rRev,
    parameter_get_utils as rParaGet,
)

from duHast.Revit.Revisions import new_revision as rNewRev

# import Autodesk
import Autodesk.Revit.DB as rdb

# tuples containing revision information
REVISION_DATA = namedtuple(
    "revisionData",
    "description issued_by issued_to revision_number_type revision_date tag_cloud_visibility",
)


def create_revision(doc, revision_data):
    """
    Creates a revision in the document.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revision_data: Named tuple containing revision data
    :type revision_data: :class:`.revisionData`

    :return:
        Result class instance.

        - Revision created status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain the message revision created successfully.
        - result.result: will contain list with single entry: the new revision created

        On exception:

        - result.status (bool) will be False.
        - result.message will contain the exception message.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()

    revit_version = rRev.get_revit_version_number(doc)

    def action():
        action_return_value = res.Result()
        if revit_version <= 2022:
            action_return_value = rNewRev.new_revision_action_2022(doc, revision_data)
        else:
            action_return_value = rNewRev.new_revision_action_2023(doc, revision_data)
        return action_return_value

    transaction = rdb.Transaction(doc, "adding revision to file")
    return_value = rTran.in_transaction(transaction, action)
    return return_value


def mark_revision_as_issued(doc, revision):
    """
    Sets a revision status to 'Issued'.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revision: The revision
    :type revision: Autodesk.REvit.DB.Revision

    :return:
        Result class instance.

        - Revision status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain the message revision marked as issued successfully.
        - result.result: empty list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain the exception message.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()

    def action():
        action_return_value = res.Result()
        try:
            revision.Issued = True
            action_return_value.update_sep(True, "Marked revision as issued.")
        except Exception as e:
            action_return_value.update_sep(
                False, "Failed to mark revision as issued with exception: " + str(e)
            )
        return action_return_value

    transaction = rdb.Transaction(doc, "Setting revision to issued")
    return_value = rTran.in_transaction(transaction, action)
    return return_value


def mark_revision_as_issued_by_revision_id(doc, revision_id):
    """
    Sets a revision, identified by its id, status to 'Issued'.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revision_id: The Id of the revision.
    :type revision_id: Autodesk.Revit.DB.ElementId

    :return:
        Result class instance.

        - Revision marked as issued status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain the message revision marked as issued successfully.
        - result.result: empty list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain the exception message.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    # get all revisions in file
    revisions_in_model = rdb.Revision.GetAllRevisionIds(doc)
    if revision_id in revisions_in_model:
        # get the revision element
        revision = doc.GetElement(revision_id)
        # mark revision as issued
        return_value = mark_revision_as_issued(doc, revision)
    else:
        return_value.update_sep(
            False, "Revision with id provided does not exist in model."
        )
    return return_value


def add_revisions_to_sheet(doc, sheet, revision_ids):
    """
    Adds revisions to single sheet

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param sheet: The sheet to add the revision to.
    :type sheet: Autodesk.Revit.DB.SheetView
    :param revision_ids: List of revision ids
    :type revision_ids: [Autodesk.Revit.ElementId]

    :return:
        Result class instance.

        - Revision adding to sheet status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain the message revision added to sheet  successfully.
        - result.result: empty list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain the exception message.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    # get revisions already on sheet (this is important, since they need to be passed in again when adding a new revision!)
    # this call converts ids to a c# List<ElementId> : ids.ToList[ElementId]()
    ids = sheet.GetAdditionalRevisionIds()
    # add new revisions to sheet
    for revision_id in revision_ids:
        ids.Add(revision_id)

    # commit new revisions in a transaction
    def action():
        action_return_value = res.Result()
        try:
            sheet.SetAdditionalRevisionIds(ids)
            action_return_value.update_sep(True, "Added revision(s) to sheet.")
        except Exception as e:
            action_return_value.update_sep(
                False, "Failed to add revision(s) to sheet with exception: " + str(e)
            )
        return action_return_value

    transaction = rdb.Transaction(doc, "adding revision to sheet")
    return_value = rTran.in_transaction(transaction, action)
    return return_value


def get_issued_revisions(doc):
    """
    Get all issued revisions in a model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: list of revision objects
    :rtype: [Autodesk.Revit.DB.Revision]
    """

    issued_revisions = []
    # get all revisions in file
    revisions_in_model = rdb.Revision.GetAllRevisionIds(doc)
    for revision_id in revisions_in_model:
        rev = doc.GetElement(revision_id)
        if rev.Issued == True:
            issued_revisions.append(rev)
    return issued_revisions


def get_last_issued_revision(doc):
    """
    Get the last issued revision from a model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A revision object or None if no issued revision in the model
    :rtype: Autodesk.Revit.DB.Revision or None
    """

    issued_revs = get_issued_revisions(doc)
    if len(issued_revs) > 0:
        issued_revs.sort(key=lambda x: x.SequenceNumber)
        return issued_revs[-1]
    else:
        return None


def change_revision_sequence_number(doc, revision, new_sequence_number):
    """
    Updates the revision sequence number of a revision.

    Pop the revision from its current index in the revision sequence list.
    Insert it again at the new index

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revision: the revision
    :type revision: Autodesk.Revit.DB.Revision
    :param new_sequence_number: The new index (sequence number ) of the revision. 1 based!!
    :type new_sequence_number: int

    :return:
        Result class instance.

        - Revision re-ordering status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain the message revision re-ordered successfully.
        - result.result: new revision sequence list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain the exception message.
    """

    return_value = res.Result()
    try:
        # note revision sequence list starts with 1 with internal revision id index list starts at 0!
        new_sequence_number = new_sequence_number - 1
        revision_id = revision.Id
        revisions_in_model = rdb.Revision.GetAllRevisionIds(doc)
        if revision_id in revisions_in_model:
            # this is a c# list: List[ElementId] hence .IndexOf
            current_index = revisions_in_model.IndexOf(revision_id)
            if new_sequence_number != current_index:
                # remove the revision at the current index
                # this is a c# list: List[ElementId] hence .RemoveAt
                revisions_in_model.RemoveAt(current_index)
                # and add at the new index
                # this is a c# list: List[ElementId] hence .Insert
                revisions_in_model.Insert(new_sequence_number, revision_id)
                # update the revision sequence in the model
                return_value.update(re_order_revisions(doc, revisions_in_model))
            else:
                return_value.update_sep(
                    True,
                    "New sequence number: {} is identical to current sequence number: {}.".format(
                        new_sequence_number, current_index
                    ),
                )
        else:
            return_value.update_sep(
                False,
                "Invalid revision provided. Id {} does not exist in model revision sequence".format(
                    revision_id
                ),
            )
    except Exception as e:
        return_value.update_sep(
            False,
            "An exception occurred in change revision sequence number: {}".format(e),
        )
    return return_value


def re_order_revisions(doc, revision_sequence):
    """
    Attempts to replace the current revision sequence with the one past in.

    Note: both sequences must contain the same revisions...just in a different order.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revision_sequence: List of revision element ids describing the new sequence.
    :type revision_sequence: IList<ElementId>
    :return:
        Result class instance.

        - Revision re-ordering status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain the message revision re-ordered successfully.
        - result.result: new revision sequence list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain the exception message.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()

    def action():
        action_return_value = res.Result()
        try:
            rdb.Revision.ReorderRevisionSequence(doc, revision_sequence)
            action_return_value.update_sep(True, "Re-ordered revisions in model.")
            action_return_value.result = rdb.Revision.GetAllRevisionIds(doc)
        except Exception as e:
            action_return_value.update_sep(
                False, "Failed to re-order revision(s) with exception: {}".format(e)
            )
        return action_return_value

    transaction = rdb.Transaction(doc, "Re-ordering revisions in model.")
    return_value = rTran.in_transaction(transaction, action)
    return return_value


# ---------------------------------------- deleting revisions --------------------------------------------


def _check_Revision_against_filters(revision, revision_description_filter):
    """
    Checks a revision against a list of revision filters and returns the outcome of all checks.

    :param revision: The revision to be checked
    :type revision: Autodesk.Revit.DB.Revision
    :param revision_description_filter: List of revision filters (refer to REVISION_KEEP_RULES at end of script)
    :type revision_description_filter: [parameter name, comparison function, comparison value]

    :return: A boolean representing the outcome of all checks performed (logical AND)
    :rtype: bool
    """

    paras = revision.GetOrderedParameters()
    rule_match = True
    for para_name, para_condition, condition_value in revision_description_filter:
        for p in paras:
            if p.Definition.Name == para_name:
                match = rParaGet.check_parameter_value(
                    p, para_condition, condition_value
                )
                rule_match = rule_match and match
    return rule_match


def delete_all_revisions_in_model(doc, revision_description_filter=[]):
    """
    Deletes all revision in file passing filter in one transaction.

    :param doc: Current model document
    :type doc: Autodesk.Revit.DB.Document
    :param revision_description_filter: list of filters, defaults to []
    :type revision_description_filter: list, optional

    :return:
        Result class instance.

        - result.status. True if all possible revisions where deleted successfully, otherwise False.
        - result.message will contain the name(s) of the revisions excluded by filter and number of revisions deleted.
        - result.result empty list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain generic exception message.
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    to_delete = []
    filtered_at_least_one = False
    revisions = rdb.FilteredElementCollector(doc).OfClass(rdb.Revision)
    for rev in revisions:
        if len(revision_description_filter) > 0:
            ruleMatch = _check_Revision_against_filters(
                rev, revision_description_filter
            )
            if ruleMatch == True:
                # delete view
                to_delete.append(rev.Id)
                return_value.append_message(
                    "Revision: {} {} will be deleted.".format(
                        rev.Description, rev.RevisionDate
                    )
                )
            else:
                filtered_at_least_one = True
                return_value.append_message(
                    "Revision: {} {} will not be deleted.".format(
                        rev.Description, rev.RevisionDate
                    )
                )
        else:
            to_delete.append(rev.Id)
            return_value.append_message(
                "Revision: {} {} will be deleted.".format(
                    rev.Description, rev.RevisionDate
                )
            )

    # check if any revisions to delete
    if len(to_delete) > 0:
        # check if at least one revision was filtered out and will therefore remain in the file
        # if not remove one manually since Revit requires at least one revision in the file
        if filtered_at_least_one == False:
            # one revision need to stay in file...
            to_delete.pop()
            return_value.append_message(
                "Removing at one revision from selection since Revit requires at least one revision in the model."
            )

        # delete them
        def action():
            action_return_value = res.Result()
            action_return_value.append_message(
                "Attempting to delete revisions: {}".format(len(to_delete))
            )
            try:
                doc.Delete(to_delete.ToList[rdb.ElementId]())
                action_return_value.append_message(
                    "Deleted {} revisions.".format(len(to_delete))
                )
            except Exception as e:
                action_return_value.update_sep(
                    False, "Failed to delete revisions with exception: {}".format(e)
                )
            return action_return_value

        transaction = rdb.Transaction(doc, "Deleting Revisions")
        return_value.update(rTran.in_transaction(transaction, action))
    else:
        return_value.update_sep(
            False, "Only one revision in file which can not be deleted!"
        )
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit sheet creation. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects.result import Result
from duHast.Revit.Common.transaction import in_transaction_with_failure_handling
from duHast.Revit.Common.parameter_set_utils import set_parameter_without_transaction_wrapper_by_name
from duHast.Revit.Common.Objects.FailureHandlingConfiguration import (
    FailureHandlingConfig,
)


from Autodesk.Revit.DB import FamilySymbol, Transaction, ViewSheet

def create_sheet(doc, sheet_number, sheet_name, title_block_family_symbol, additional_properties={}):
    """
    Creates a revit sheet.

    :param doc: The Revit document.
    :type doc: Document
    :param sheet_number: The new sheet number
    :type sheet_number: str
    :param sheet_name: The new sheet name
    :type sheet_name: str
    :param title_block_family_symbol: The title block type to be used
    :type title_block_family_symbol: Autodesk.Revit.DB.FamilySymbol
    :param additional_properties: a dictionary representing parameter names on the new sheet and their value, defaults to {}
    :type additional_properties: dict, optional
    :raises ValueError: If sheet number is not a string
    :raises ValueError: If sheet name is not a string
    :raises ValueError: If title_block_family_symbol is not a Autodesk.Revit.DB.FamilySymbol
    :raises ValueError: If sheet failed to create

    :return: Result class instance.

            - .status True if sheet was created and parameters where updated successfully. Otherwise False.
            - .message will contain stats.
            - .result will be a list containing the sheet instance created.

    :rtype: :class:`.Result`

    """
    return_value = Result()

    try:
        # type checking
        if not(isinstance (sheet_number, str)):
            raise ValueError ("sheet number needs to be of type str, got {} instead.".format(type(sheet_number)))
        
        if not(isinstance(sheet_name, str)):
            raise ValueError ("sheet name needs to be of type str, got {} instead.".format(type(sheet_name)))
        
        if (not(isinstance(title_block_family_symbol, FamilySymbol))):
            raise ValueError ("title_block_family_symbol needs to be of type Autodesk.Revit.DB.FamilySymbol, got {} instead.".format(type(title_block_family_symbol)))
        
        # set up an action which creates the sheet and assigns properties
        def action():
            action_return_value = Result()
            try:
                view_sheet = ViewSheet.Create(doc, title_block_family_symbol.Id)

                # check if sheet was actually created?
                if not view_sheet:
                    raise ValueError("Failed to create new ViewSheet.")
                
                # set sheet standard properties
                view_sheet.SheetNumber = sheet_number
                view_sheet.Name = sheet_name

                # set additional sheet parameter values if required
                for key,value in additional_properties.items():
                    # convert value to string if not already
                    new_value = value
                    if not(isinstance(new_value, str)):
                        new_value=str(new_value)
                    # set the new value
                    set_parameter_status = set_parameter_without_transaction_wrapper_by_name(element=view_sheet, parameter_name=key,parameter_value=new_value)
                    # update the return value
                    action_return_value.update(set_parameter_status)
                
                action_return_value.append_message("Successfully created sheet {} {}".format(sheet_number, sheet_name))
                action_return_value.result.append(view_sheet)
            except Exception as e:
                action_return_value.update_sep(False,"{}".format(e))
            return action_return_value
        
        # define failure handling for the transaction ( roll back on any warnings or errors )
        failure_handling_settings = FailureHandlingConfig(
            roll_back_on_warning=True,
            print_warnings=False,
            roll_back_on_error=True,
            print_errors=False,
        )

        # set up the transaction
        trans = Transaction(doc, "creating sheet {} {}".format(sheet_number, sheet_name))

        # execute the transaction with failure handling
        result_transaction = in_transaction_with_failure_handling(
            transaction=trans,
            action=action,
            failure_config=failure_handling_settings,
        )

        # update the return value with the result of the transaction
        return_value.update(result_transaction)
        
    except Exception as e:
        return_value.update_sep(False, "failed to create sheet with exception: {}".format(e))

    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a Revit rooms export to DATA class functions. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from Autodesk.Revit.DB import BuiltInParameter, Element

from duHast.Revit.Common import (
    parameter_get_utils as rParaGet,
    phases as rPhase,
)
from duHast.Data.Objects.Collectors import data_room as dRoom
from duHast.Revit.Common.Geometry import to_data_conversion as rGeo
from duHast.Revit.Rooms.rooms import get_all_rooms
from duHast.Revit.Rooms.Geometry.geometry import get_2d_points_from_revit_room
from duHast.Revit.Exports.export_data import (
    get_model_data,
    get_instance_properties,
    get_design_set_data,
)

from duHast.Utilities.utility import encode_utf8


def populate_data_room_object(doc, revit_room):
    """
    Returns a custom room data objects populated with some data from the revit model room past in.
    data points:
    - room name, number, id
    - if exists: parameter value of SP_Room_Function_Number
    - level name and id (if not placed 'no level' and -1)
    - Design set and option
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revit_room: The room.
    :type revit_room: Autodesk.Revit.DB.Architecture.Room

    :return: A room data instance.
    :rtype: :class:`.DataRoom`
    """

    # set up data class object
    data_r = dRoom.DataRoom()
    # get room geometry (boundary points)
    revit_geometry_point_groups = get_2d_points_from_revit_room(revit_room)
    if len(revit_geometry_point_groups) > 0:
        room_point_groups_as_doubles = []
        for room_point_group_by_poly in revit_geometry_point_groups:
            data_geometry_converted = rGeo.convert_xyz_in_data_geometry_polygons(
                doc, room_point_group_by_poly
            )
            room_point_groups_as_doubles.append(data_geometry_converted)
        data_r.polygon = room_point_groups_as_doubles
        
        # get design set data
        design_set = get_design_set_data(doc=doc, element=revit_room)
        data_r.design_set_and_option = design_set

        # get instance properties
        instance_props = get_instance_properties(revit_room)
        data_r.instance_properties = instance_props

        # get the model name
        model = get_model_data(doc=doc)
        data_r.revit_model = model

        # get phase name
        data_r.phasing.created = encode_utf8(
            rPhase.get_phase_name_by_id(
                doc,
                rParaGet.get_built_in_parameter_value(
                    revit_room,
                    BuiltInParameter.ROOM_PHASE,
                    rParaGet.get_parameter_value_as_element_id,
                ),
            )
        )
        data_r.phasing.demolished = -1

        # get level data
        try:
            data_r.level.name = encode_utf8(Element.Name.GetValue(revit_room.Level))
            data_r.level.id = revit_room.Level.Id.IntegerValue
        except:
            data_r.level.name = "no level"
            data_r.level.id = -1
        return data_r

    else:
        return None


def get_all_room_data(doc):
    """
    Returns a list of room data objects for each room in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of room data instances.
    :rtype: list of  :class:`.DataRoom`
    """

    all_room_data = []
    rooms = get_all_rooms(doc)
    for room in rooms:
        rd = populate_data_room_object(doc, room)
        if rd is not None:
            all_room_data.append(rd)
    return all_room_data



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Data sheet size class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import json
from duHast.Utilities.Objects.base import Base
from duHast.Data.Objects.Collectors.Properties.data_property_names import (
    DataPropertyNames,
)


class DataSheetSize(Base):
    data_type = "sheet_size"

    def __init__(self, name=None, width=0.0, height=0.0, j=None,**kwargs):
        """
        Class constructor

        :param j:  json formatted dictionary of this class, defaults to {}
        :type j: dict, optional
        """

        # store data type  in base class
        super(DataSheetSize, self).__init__(**kwargs)

        json_var = None
        # check if any json data was past in with constructor!
        if j is not None:
            # check type of data that came in:
            if isinstance(j, str):
                # a string
                json_var = json.loads(j)
            elif isinstance(j, dict):
                # no action required
                json_var = j.copy()
            else:
                raise TypeError(
                    "Argument j supplied must be of type string or type dictionary. Got {} instead.".format(
                        type(j)
                    )
                )

            # attempt to populate from json
            try:
                self.name = json_var.get(DataPropertyNames.NAME, None)
                if not (isinstance(self.name, str)):
                    raise ValueError(
                        "name needs to be of type str, got {} instead.".format(
                            type(self.name)
                        )
                    )

                self.width = json_var.get(DataPropertyNames.WIDTH, 0.0)
                if not (isinstance(self.width, float)):
                    raise ValueError(
                        "width needs to be of type float, got {} instead.".format(
                            type(self.name)
                        )
                    )

                self.height = json_var.get(DataPropertyNames.HEIGHT, 0.0)
                if not (isinstance(self.height, float)):
                    raise ValueError(
                        "height needs to be of type float, got {} instead.".format(
                            type(self.name)
                        )
                    )
            except Exception as e:
                raise type(e)(
                    "Node {} failed to initialise with: {}".format(self.data_type, e)
                )
        else:
            # set default values as past in only if there is no json
            self.name = name
            if not (isinstance(self.name, str)):
                raise ValueError(
                    "name needs to be of type str, got {} instead.".format(type(self.name))
                )
            

            self.width = width
            if not (isinstance(self.width, float)):
                raise ValueError(
                    "width needs to be of type float, got {} instead.".format(
                        type(self.name)
                    )
                )
            self.height = height
            if not (isinstance(self.height, float)):
                raise ValueError(
                    "height needs to be of type float, got {} instead.".format(
                        type(self.name)
                    )
                )

    def is_matching_size(self, width, height):
        """
        Check if supplied values are within a 20mm band of the actual paper size

        Args:
            width (_type_): _description_
            height (_type_): _description_

        Returns:
            _type_: _description_
        """
        # check if the past in values are within 10mm band of the sheet size

        if (
            width
            <= self.width
            + 10  # supplied width can be a max of 10 mm larger than the paper format
            and width
            >= self.width
            - 10  # supplied width can be a max of 10 mm smaller than the paper format
            and height
            <= self.height
            + 10  # supplied height can be a max of 10 mm larger than the paper format
            and height
            >= self.height
            - 10  # supplied height can be a max of 10 mm smaller than the paper format
        ):
            return True
        else:
            return False

    def __eq__(self, other):
        if not isinstance(other, DataSheetSize):
            return NotImplemented

        return (
            self.name == other.name
            and self.width == other.width
            and self.height == other.height
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash((self.name, self.width, self.height))



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A decorator factory that returns a decorator that can be used to add logging to a function
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Peter Smith
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import functools
import traceback
from duHast.Utilities.benchmarking import measure_time_wrapper


def get_add_logger_decorator(
    log_obj_inst, errors_in_console=True, suppress_exceptions=True, record_time=False
):
    """
    This is a decorator factory that returns a decorator that can be used to add logging to a function.
    Pass this function an instance of the LoggerObject class from duHast.Utilities.Objects.logger_object.py.
    Inherit and extend this class to the needs of your application. This will connect the decorator to
    the logger object application wide. Assign the return value to a variable called log or add_logger etc.

    :param log_obj_inst: An instance of the LoggerObject class from duHast.Utilities.Objects.logger_object.py
    :type log_obj_inst: duHast.Utilities.Objects.logger_object.LoggerObject
    :param errors_in_console: Whether to output errors to the console. Defaults to True. Turn off to handle in the application code
    :type errors_in_console: bool
    :return: A decorator that can be used to add logging to a function
    :rtype: function

    """

    def add_logger(
        log_level=(10, 30),
        measure_time=record_time,
    ):
        """
        Decorator function to add logging to a function.
        By default it will log the start and end of the function, the arguments passed in and the return value.
        It will also log the elapsed time if measure_time is set to True.

        Use it like:
        @add_logger()
        def example_function():
            do_stuff()

        :param log_level: A tuple of the python log levels for the file and console handlers. See
        https://docs.python.org/3/library/logging.html#levels for log levels
        :type log_level: tuple/list of int (file_log_level, console_log_level)
        :param measure_time: Whether to measure and log the time the function takes to execute
        :type measure_time: bool
        :return: The decorated function
        :rtype: function
        """

        def decorator_log(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                func_name = func.__name__
                logger_in = log_obj_inst
                # First check to see if the log levels passed in are different and update
                cur_file_log_level = logger_in.file_handler.level
                cur_console_log_level = logger_in.console_handler.level
                updated_levels = False
                if (
                    cur_file_log_level != log_level[0]
                    or cur_console_log_level != log_level[1]
                ):
                    # Update the log levels
                    logger = logger_in.update_log_level(log_level)
                    logger.info("Updated log levels to {}".format(log_level))
                    updated_levels = True
                try:
                    # Log the start of the function execution
                    logger = logger_in.get_logger_obj()
                    logger.info(
                        "------------ Started function: {}. Log level set at: {}".format(
                            func_name, log_level
                        )
                    )
                    # Get the args passed into the function to log
                    args_repr = [repr(a) for a in args]
                    kwargs_repr = [
                        "{}={}".format(k, repr(v)) for k, v in kwargs.items()
                    ]
                    # Log those arguments
                    if len(args_repr + kwargs_repr) > 0:
                        signature = ", ".join(args_repr + kwargs_repr)
                        logger.debug(
                            "function {} called with args: {}".format(
                                func_name, signature
                            )
                        )
                    else:
                        logger.info(
                            "function {} called with no arguments".format(func_name)
                        )

                except Exception as e:
                    pass

                result = None
                error = False

                try:
                    # Execute the wrapped function either with or without time measurement
                    if measure_time:
                        result = measure_time_wrapper(
                            func, logger, measure=measure_time
                        )(*args, **kwargs)

                        return result
                    else:
                        result = func(*args, **kwargs)
                        return result

                except Exception as e:
                    tb = traceback.format_exc().strip().split("\n")

                    log_msg = "Error raised in {} function. {}\n\n{}".format(
                        func_name, tb[-1], "\n".join(tb)
                    )
                    if errors_in_console:
                        logger.error(log_msg)

                    else:
                        logger.error(log_msg, extra={"block": "console"})

                    if not suppress_exceptions:
                        raise e
                    error = True
                finally:
                    # Log the return value
                    if error:
                        # TODO: add support message
                        logger.info(
                            "Finished function {} with errors.".format(func_name)
                        )
                    else:
                        if result:
                            logger.info(
                                "Finished function: {}. Return value is: {}".format(
                                    func_name, repr(result)
                                )
                            )
                        else:
                            logger.info("Finished function: {}".format(func_name))
                    # Reset the log levels if they were updated
                    if updated_levels:
                        logger = logger_in.update_log_level(
                            (cur_file_log_level, cur_console_log_level)
                        )

            return wrapper

        return decorator_log

    return add_logger



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of functions around Extensible storage deletion in Revit.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Notes from the building coder website: https://thebuildingcoder.typepad.com/blog/2022/11/extensible-storage-schema-deletion.html


The devteam replied: After running DeleteSchemas macro, open Manage > Purge Unused. In the tree, select Extensible Storage Schema. Check the schema a9dc2b48 and click OK to purge it. Run ListSchemas – the schema is purged. So, please use Purge Unused to delete schemas without entities.

Extensible storage schema is an application-wide object. If it exists at all in the application, it will populate and "infect" every single document that you touch. That makes it hard to remove, and complicated to understand.


Personally I found:

- The schema is not removed from the document when you delete it. You have to purge it manually.
- If the schema exists in a link and you delete it, it will not go away ... unload  any links or replace any links with dummy links without the schema and attempt to delete and purge again.
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#



from duHast.Utilities.Objects.result import Result

from System import Guid

from Autodesk.Revit.DB import Transaction
from Autodesk.Revit.DB.ExtensibleStorage import  Schema

from duHast.Revit.Common.transaction import in_transaction

def delete_schema_by_guid(doc, guid, transaction_manager = in_transaction):
	"""
	Deletes a schema by its guid. This will also delete all entities that are using this schema.

	Note:


	:param doc: The Revit document
	:type doc: Autodesk.Revit.DB.Document
	:param guid: The guid of the schema to delete
	:type guid: str
	:param transaction_manager: The transaction manager to use. Defaults to in_transaction.
	:type transaction_manager: function
	:return: A Result object with the result of the operation
	:rtype: Result
	"""

	return_value = Result()
	try:
		
		schemas = Schema.ListSchemas()
	
		# define an action to be executed in a transaction
		def action ():
			action_return_value = Result()

			found_match = False
			for schema in schemas:
				if schema.GUID.ToString() == guid:
					action_return_value.append_message("Schema found")
					found_match = True
					doc.EraseSchemaAndAllEntities(schema)
					action_return_value.append_message("Schema erased")
					# there should only be one schema with this guid...get out of the loop
					return action_return_value
			
			#  check if we found a match
			if not found_match:
				action_return_value.append_message("Schema not found")
				
				return action_return_value
		
		if transaction_manager:
			# run the action in a transaction
			tranny = Transaction(doc, "Erase EStorage: {}".format(guid))
			return_value = transaction_manager(tranny, action)
		
		else:
			# run the action without a transaction
			return_value = action()

		return return_value
	except Exception as e:
		return_value.update_sep(False, str(e))
		return return_value




"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit sheet to data sheet conversion. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Exports.export_data import get_instance_properties
from duHast.Revit.Views.views import get_viewport_on_sheets
from duHast.Revit.Views.Export.view_ports_to_data import (
    convert_revit_viewport_to_data_instance,
    convert_revit_schedule_sheet_instances_to_data_instance,
)
from duHast.Data.Objects.Collectors.data_sheet import DataSheet
from duHast.Revit.Views.sheets import get_sheets_by_filters, get_title_block_from_sheet
from duHast.Revit.Views.schedules import get_schedule_instance_on_sheet
from duHast.Revit.Common.Geometry.to_data_conversion import convert_revit_bounding_box_to_geometry2_bounding_box
from duHast.Data.Objects.Collectors.Properties.data_sheet_size_names import DataSheetSizeNames


def get_title_block_size(doc, sheet):
    
    title_block_instance = get_title_block_from_sheet(doc=doc, sheet=sheet)
    if title_block_instance is None:
        return None
    
    # get a 2D bounding box from the title block instance
    bbox = convert_revit_bounding_box_to_geometry2_bounding_box(bounding_box=title_block_instance.BoundingBox())
    
    # get supported sheet sizes
    sheet_sizes_names= DataSheetSizeNames()
    sizes = sheet_sizes_names.get_all_supported_sizes()
    
    # find a match
    for supported_size in sizes:
        if(supported_size.is_matching_size(width=bbox.width, height=bbox.depth)):
            return supported_size.name
    
    # uh no match found...
    return None
    
    
def convert_revit_sheet(doc, sheet):
    """
    Convertes a revit sheet to a data instance sheet.

    :param doc: The Revit document.
    :type doc: Autodesk.Revit.DB.Document
    :param sheet: A revit sheet
    :type sheet: Autodesk.Revit.DB.view

    :return: A data sheet instance
    :rtype: :class:`.DataSheet`
    """

    # instantiate new data sheet object
    data_sheet = DataSheet()
    
    # get the sheet size
    

    # get any instance parameters properties
    instance_properties = get_instance_properties(sheet)
    data_sheet.instance_properties = instance_properties

    # get view ports on sheet
    revit_view_ports = get_viewport_on_sheets(doc=doc, sheets=[sheet])

    # convert to data objects
    view_ports_converted = []
    for revit_view_port in revit_view_ports:
        view_port_data = convert_revit_viewport_to_data_instance(
            doc=doc, revit_view_port=revit_view_port
        )

        # check if this is a view port of interest, otherwise ignore
        if view_port_data:
            view_ports_converted.append(view_port_data)

    # get schedule instances on sheet
    schedule_sheet_instances = get_schedule_instance_on_sheet(doc, sheet)
    if schedule_sheet_instances and len(schedule_sheet_instances) > 0:
        instances_converted = convert_revit_schedule_sheet_instances_to_data_instance(
            doc=doc,
            sheet=sheet,
            revit_schedule_sheet_instances=schedule_sheet_instances,
        )
        view_ports_converted = view_ports_converted + instances_converted

    # add them to sheet
    data_sheet.view_ports = view_ports_converted
    return data_sheet


def get_all_sheet_data(doc):
    """
    Gets a list of sheet data objects for each sheet element in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of data sheet instances.
    :rtype: list of :class:`.DataSheet`
    """

    all_sheet_data = []
    sheets = get_sheets_by_filters(doc=doc, view_rules=None)
    for sheet in sheets:
        sd = convert_revit_sheet(doc=doc, sheet=sheet)
        if sd is not None:
            all_sheet_data.append(sd)
    return all_sheet_data



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to batch processor.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


def adjust_session_id_for_file_name(id):
    """
    Removes chevrons and replace colons with underscores in session id supplied by revit batch processor so it\
        can be used in a file name.

    :param id: Session id supplied by revit batch processor.
    :type id: str

    :return: Re-formatted session id.
    :rtype: str
    """

    # remove colons
    session_id_changed = id.replace(":", "_")
    # remove chevrons
    session_id_changed = session_id_changed[1:-1]
    return session_id_changed


def adjust_session_id_file_name_back(file_name_id):
    """
    Re-introduces chevrons and replaces underscores with colons to match session Id format used in batch processor to a\
        file name using a batch processor supplied id.

    :param file_name_id: A file name containing a session id with all illegal characters replaced.
    :type file_name_id: str

    :return: A session id.
    :rtype: str
    """

    # re-instate colons
    session_id_changed = file_name_id.replace("_", ":")
    # remove chevrons
    session_id_changed = "<" + session_id_changed + ">"
    return session_id_changed


def adjust_session_id_for_directory_name(id):
    """
    Removes chevrons and replace colons, full stops, dashes with underscores in session id supplied by revit batch processor so it\
        can be used in a folder name.

    :param id: Session id supplied by revit batch processor.
    :type id: str

    :return: Re-formatted session id.
    :rtype: str
    """

    # remove colons
    session_id_changed = id.replace(":", "_")
    # remove spaces
    session_id_changed = session_id_changed.replace(" ", "_")
    # remove full stops
    session_id_changed = session_id_changed.replace(".", "_")
    # remove dashes
    session_id_changed = session_id_changed.replace("-", "_")
    # remove chevrons
    session_id_changed = session_id_changed.replace("<", "")
    session_id_changed = session_id_changed.replace(">", "")
    return session_id_changed



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit levels helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import System


clr.AddReference("System.Core")
from System import Linq

clr.ImportExtensions(Linq)

# import common library modules
from duHast.Revit.Common.common import get_ids_from_element_collector

from duHast.Utilities.unit_conversion import convert_imperial_feet_to_metric_mm

# import Autodesk
from Autodesk.Revit.DB import (
    BuiltInCategory,
    ElementCategoryFilter,
    FamilySymbol,
    FilteredElementCollector,
    Level,
)


def get_levels_in_model(doc):
    """
    Get all levels in model

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A collector with all levels in model.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = FilteredElementCollector(doc).OfClass(Level)
    return collector


def get_levels_in_view(doc, view):
    """
    Get all levels in a view

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document
    :param view: The view to get grids from.
    :type view: Autodesk.Revit.DB.View
    :return: A collector with all grids in view.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = FilteredElementCollector(doc, view.Id).OfClass(Level)
    return collector


# --------------------------------------------- utility functions ------------------


def get_levels_list_ascending(doc):
    """
    Gets a filtered element collector of all levels in the model ascending by project elevation.

    Filters by category.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector of levels
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = (
        FilteredElementCollector(doc)
        .OfCategory(BuiltInCategory.OST_Levels)
        .WhereElementIsNotElementType()
        .ToList()
        .OrderBy(lambda l: l.ProjectElevation)
    )
    return collector


def get_levels_list_descending(doc):
    """
    Gets a filtered element collector of all levels in the model descending by project elevation.

    Filters by category.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector of levels
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    levels_ascending = get_levels_list_ascending(doc)
    # using .Reverse() since this is a c# list, to get descending order rather than .reverse() which is a python list method
    levels_descending = levels_ascending.Reverse()
    return levels_descending


def get_level_elevation_by_name(doc, level_name):
    level_elevation = None
    levels = get_levels_in_model(doc=doc)
    for level in levels:
        if level.Name == level_name:
            return level.Elevation
    return level_elevation


# ------------------------------------------------- filters --------------------------------------------------------------------


def get_all_level_heads_by_category(doc):
    """
    Gets a filtered element collector of all level head types in the model.

    Filters by category.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector of level heads
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = (
        FilteredElementCollector(doc)
        .OfCategory(BuiltInCategory.OST_LevelHeads)
        .WhereElementIsElementType()
    )
    return collector


def get_all_level_types_by_category(doc):
    """
    Gets a filtered element collector of all level types in the model.

    Filters by category.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector of level types.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = (
        FilteredElementCollector(doc)
        .OfCategory(BuiltInCategory.OST_Levels)
        .WhereElementIsElementType()
    )
    return collector


def get_all_level_type_ids_by_category(doc):
    """
    Gets a list of all level type ids in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A list of all level type ids.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    collector = (
        FilteredElementCollector(doc)
        .OfCategory(BuiltInCategory.OST_Levels)
        .WhereElementIsElementType()
    )
    ids = get_ids_from_element_collector(collector)
    return ids


# -------------------------------------------------  purge --------------------------------------------------------------------


def get_all_level_head_family_type_ids(doc):
    """
    Gets ids of all level head family symbols (types) in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A list of symbol ids.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = []
    filter = ElementCategoryFilter(BuiltInCategory.OST_LevelHeads)
    col = FilteredElementCollector(doc).OfClass(FamilySymbol).WherePasses(filter)
    ids = get_ids_from_element_collector(col)
    return ids


def get_nearest_level_absolute(z, levels, ignore_level_names):
    """
    Returns the nearest level matching the Z value and the offset adjustment from the level (in metric!!) in a tuple.

    Note:
    Will return None if the placement level could not be determined.
    Nearest level is the level with the shortest absolute distance to the Z value and which is not in the ignore_level_names list.
    If the Z value is below the lowest level elevation in the model the lowest level will be returned.
    If the Z value is above the highest level elevation in the model the highest level will be returned.

    :param z: The Z coordinate (metric!)
    :type z: float
    :param levels: List of levels in the model. (sorted ascending)
    :type levels: list
    :return: Returns the level and the offset adjustment (in metric) from the source family instance level.
    :rtype: tuple
    """

    # distance placeholder
    absolute_distance_to_level = None
    nearest_level = None
    for level in levels:
        # check if level is to be ignored
        if level.Name in ignore_level_names:
            continue
        # get the absolute distance
        distance_to_level = abs(z - convert_imperial_feet_to_metric_mm(level.Elevation))
        # store the first level as the nearest level
        if absolute_distance_to_level is None:
            absolute_distance_to_level = distance_to_level
            nearest_level = level
        # check if the distance is shorter than the current shortest distance
        if distance_to_level < absolute_distance_to_level:
            absolute_distance_to_level = distance_to_level
            nearest_level = level

    if nearest_level == None:
        return None, None
    else:
        offset_from_level = z - convert_imperial_feet_to_metric_mm(
            nearest_level.Elevation
        )
        return nearest_level, offset_from_level


def get_nearest_lowest_level(z, levels, ignore_level_names):
    """
    Returns the nearest lowest level below the Z value and the offset adjustment from the level (in metric) in a tuple.

    Note:
    Will return None if the placement level could not be determined.
    Nearest level is the level below or at the Z value and which is not in the ignore_level_names list.

    :param z: The Z coordinate (metric!)
    :type z: float
    :param levels: List of levels in the model. (needs to be sorted ascending)
    :type levels: list
    :return: Returns the level and the offset adjustment (in metric) from the source family instance level.
    :rtype: tuple
    """

    # get level by Z coordinate of family
    nearest_level = None
    for level in levels:
        # check if level is to be ignored
        if level.Name in ignore_level_names:
            continue
        # store the lowest level for later
        if nearest_level == None:
            nearest_level = level
        # get the level by height
        if round(convert_imperial_feet_to_metric_mm(level.Elevation), 2) <= round(z, 2):
            nearest_level = level
        else:
            break

    # check if matching level was found and calc any offset
    if nearest_level != None:
        offset_from_level_adjustment = z - convert_imperial_feet_to_metric_mm(
            nearest_level.Elevation
        )
        return nearest_level, offset_from_level_adjustment
    else:
        return None, None



import logging
import datetime


def level_time_utc_msg_formatter():
    """Standard output formatting for console and file handlers"""
    # Format the date and time
    utc_now = datetime.datetime.utcnow()
    return logging.Formatter("%(levelname)s | {} | %(message)s".format(utc_now))


def level_time_local_msg_formatter():
    return logging.Formatter(
        "%(levelname)s | %(asctime)s.%(msecs)03d | %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )


def msg_only():
    """Output formatting for presentation"""
    return logging.Formatter("%(message)s")



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions to convert settings.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import json
from duHast.Revit.Exports.Objects.export_file_name_rule import ExportFileNameRule

def convert_settings_json_string_to_settings_objects(settings_string):
    """
    Convert a settings string to a list of settings objects.

    :param settings_string: The settings string to convert.
    :type settings_string: str
    :return: A list of settings objects.
    :rtype: list
    """

    # the settings string will contain a list of settings objects
    # convert the string to a list of dictionaries
    settings_list = json.loads(settings_string)

    # create a list of settings objects
    settings_objects = []
    for settings in settings_list:
        # create a settings object
        settings_object = ExportFileNameRule(settings)
        # add the settings object to the list
        settings_objects.append(settings_object)

    # return the list of settings objects
    return settings_objects


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of functions around Revit detail lines.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import math

from duHast.Utilities.Objects import result as res

# import Autodesk
from Autodesk.Revit.DB import (
    ElementTransformUtils,
    Line,
    Transform,
    XYZ,
)


def draw_detail_line_at_90_degree(doc, view, start_point, end_point):
    """
        Draws a detail line at 90 degree angle to given points, rotated around 0 end point.

        Needs to run in a transaction.

        :param doc: The current model document
        :type doc: Autodesk.Revit.DB.Document
        :param view: The plan view to draw the detail line on.
        :type view: Autodesk.Revit.DB.View
        :param start_point: Start point of line.
        :type start_point: Autodesk.Revit.DB.XYZ
        :param end_point: End point of line.
        :type end_point: Autodesk.Revit.DB.XYZ

        :return:
        Result class instance.

        - result.status. True if line was rotated and drawn on view successfully, otherwise False.
        - result.message will contain the original points and rotated start and end points.
        - result.result empty list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain generic exception message.
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    line_to_draw = Line.CreateBound(start_point, end_point)
    return_value = res.Result()
    try:
        dc = doc.Create.NewDetailCurve(view, line_to_draw)
        dc_curve = dc.GeometryCurve
        aa = dc_curve.GetEndPoint(0)
        cc = XYZ(aa.X, aa.Y, aa.Z + 10)
        axis = Line.CreateBound(aa, cc)
        ElementTransformUtils.RotateElement(doc, dc.Id, axis, math.pi / 2.0)

        dc_curve_rotated = dc.GeometryCurve

        # get a line through origin at 90 degree
        trans_form = Transform.CreateRotationAtPoint(
            XYZ.BasisZ, math.pi / 2.0, start_point
        )
        transformed_line = line_to_draw.CreateTransformed(trans_form)

        return_value.append_message(
            "transformed line origin {} and end point {}".format(
                transformed_line.GetEndPoint(0), transformed_line.GetEndPoint(1)
            )
        )

        return_value.append_message(
            "rotated line origin {} and end point {}".format(
                dc_curve_rotated.GetEndPoint(0), dc_curve_rotated.GetEndPoint(1)
            )
        )

        return_value.update_sep(
            True,
            "Drew line between origin:{} and second point: {} in view {}".format(
                start_point, end_point, view.Name
            ),
        )
    except Exception as e:
        return_value.update_sep(False, e.message)

    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit family instance translation functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from math import pi

# class used for stats reporting
from duHast.Utilities.Objects import result as res
from duHast.Revit.Common.transaction import in_transaction

from Autodesk.Revit.DB import Element, ElementTransformUtils, Line,  LocationPoint, Transaction, XYZ

# Constants
RADIAN_ANGLE_45DEGREES = pi / 4
RADIAN_ANGLE_90DEGREES = pi / 2


def get_family_location_point(element):
    """
    Returns the location point of the family instance.

    :param element: The family instance to get the location point of.
    :type element: Autodesk.Revit.DB.FamilyInstance
    :return: The location point of the family instance.
    :rtype: Autodesk.Revit.DB.XYZ
    """

    if element is not None and element.Location is not None:
        location_point = element.Location
        if isinstance(location_point, LocationPoint):
            return location_point.Point
    return None


def get_element_location_curve(element):
    """
    Returns the location curve of the element.

    :param element: The element to get the location curve of.
    :type element: Autodesk.Revit.DB.Element
    :return: The location curve of the element.
    :rtype: Autodesk.Revit.DB.Curve
    """

    location_curve = None

    if element.Location is not None:
        location_curve = element.Location.Curve

    return location_curve


def rotate_around_transform_origin(element, angle, transaction_manager =None):
    """
    Rotates an element around its origin.

    :param element: The element to rotate.
    :type element: Autodesk.Revit.DB.Element
    :param angle: The angle to rotate the element by in radians.
    :type angle: float
    """

    return_value = res.Result()

    # Get the element's transform
    transform = element.GetTotalTransform()

    # Create a line from the transform's origin and basis Z
    line = Line.CreateUnbound(transform.Origin, transform.BasisZ)

    try:
        def action():
            try:
                action_return_value = res.Result()
                ElementTransformUtils.RotateElement(element.Document, element.Id, line, angle)
                return_value.append_message( "Element rotated around origin.") 
            except Exception as e:
                action_return_value.update_sep(False, "Failed to rotate element around origin. {}".format(e))
            return action_return_value
       
        if transaction_manager is None:
            # assume there is an transaction already going on
            return_value = action()
        else:
            # create a transaction
            transaction = Transaction(element.Document, "rotated element around origin: {}".format(Element.Name.GetValue(element)))
            return_value = in_transaction(transaction, action)

    except Exception as e:
        return_value.update_sep(False, "Failed to rotate element around origin. {}".format(e))


    return return_value


def rotate_around_origin(element, angle, transaction_manager =None):
    """
    Rotates an element around its origin.

    :param element: The element to rotate.
    :type element: Autodesk.Revit.DB.Element
    :param angle: The angle to rotate the element by in radians.
    :type angle: float
    """

    return_value = res.Result()

    # get the family location point
    location_point = get_family_location_point(element)

    # Create a line from the transform's origin and basis Z
    line = Line.CreateBound(location_point, XYZ(location_point.X, location_point.Y,  location_point.Z +1000))

    try:
        def action():
            try:
                action_return_value = res.Result()
                ElementTransformUtils.RotateElement(element.Document, element.Id, line, angle)
                return_value.append_message( "Element rotated around origin.") 
            except Exception as e:
                action_return_value.update_sep(False, "Failed to rotate element around origin. {}".format(e))
            return action_return_value
       
        if transaction_manager is None:
            # assume there is an transaction already going on
            return_value = action()
        else:
            # create a transaction
            transaction = Transaction(element.Document, "rotated element around origin: {}".format(Element.Name.GetValue(element)))
            return_value = transaction_manager(transaction, action)

    except Exception as e:
        return_value.update_sep(False, "Failed to rotate element around origin. {}".format(e))


    return return_value


def move_from_point_to_point(element, start_point, end_point, transaction_manager =None):
    """
    Moves an element from a start point to an end point. (assume same Z value)

    :param element: The element to move.
    :type element: Autodesk.Revit.DB.Element
    :param start_point: The start point of the move.
    :type start_point: Autodesk.Revit.DB.XYZ
    :param end_point: The end point of the move.
    :type end_point: Autodesk.Revit.DB.XYZ
    """

    return_value = res.Result()

    try:
        def action():
            try:
                action_return_value = res.Result()
                ElementTransformUtils.MoveElement(element.Document, element.Id, end_point - start_point)
                return_value.append_message( "Element moved from point to point.") 
            except Exception as e:
                action_return_value.update_sep(False, "Failed to move element from point to point. {}".format(e))
            return action_return_value
       
        if transaction_manager is None:
            # assume there is an transaction already going on
            return_value = action()
        else:
            # create a transaction
            transaction = Transaction(element.Document, "moved element from point to point: {}".format(Element.Name.GetValue(element)))
            return_value = transaction_manager(transaction, action)

    except Exception as e:
        return_value.update_sep(False, "Failed to move element from point to point. {}".format(e))

    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit view schedules. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from Autodesk.Revit.DB import (
    BuiltInCategory,
    Category,
    FilteredElementCollector,
    FilteredElementCollector, 
    ScheduleSheetInstance,
    ViewSchedule,
    ViewType
    )

from duHast.Revit.Views.views import get_views_of_type


def get_schedule_ids_on_sheets(doc):
    """
    Gets view ids of all schedules with instances placed on a sheet
    
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List containing schedule Id's.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = []
    col = FilteredElementCollector(doc).OfClass(ScheduleSheetInstance)
    for s in col:
        if s.ScheduleId not in ids:
            ids.append(s.ScheduleId)
    return ids


def get_schedules_not_on_sheets(doc):
    """
    Gets all schedules without an instance placed on a sheet.
    
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: list of schedules without a sheet schedule instance.
    :rtype: list of Autodesk.Revit.DB.View
    """

    schedules_not_on_sheets = []
    # get schedules on sheets
    ids_on_sheets = get_schedule_ids_on_sheets(doc)
    # get all schedules in model
    schedules_in_model = get_views_of_type(doc, ViewType.Schedule)
    # loop and filter out schedules not on sheets
    for schedule in schedules_in_model:
        if schedule.Id not in ids_on_sheets:
            schedules_not_on_sheets.append(schedule)
    return schedules_not_on_sheets


def get_schedule_instance_on_sheet(doc, sheet):
    """
    Returns a list containing all schedule sheet instances on a sheet.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param sheet: A sheet
    :type sheet: Autodesk.Revit.DB.ViewSheet
    :return: a list of schedule sheet instances or empty list if none found
    :rtype: [Autodesk.revit.DB.ScheduleSheetInstance]
    """

    schedule_instances_on_sheet = []
    col = FilteredElementCollector(doc).OfClass(ScheduleSheetInstance)

    # Filter the instances where the OwnerViewId matches the specified sheet_id
    for schedule_sheet_instance in col:
        if schedule_sheet_instance.OwnerViewId == sheet.Id:
            # rule oout revision schedules...
            schedule = doc.GetElement(schedule_sheet_instance.ScheduleId)
            if not schedule.IsTitleblockRevisionSchedule:
                schedule_instances_on_sheet.append(schedule_sheet_instance)
    
    return schedule_instances_on_sheet


def get_schedules(doc):
    """
    Get all schedules in the current document excluding titleblock revision schedules.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of schedules in the current document.
    :rtype: list[Autodesk.Revit.DB.ViewSchedule]
    """
    schedules_filtered = []
    collector = FilteredElementCollector(doc).OfClass(ViewSchedule)
    
    for schedule in collector:
        if not schedule.IsTitleblockRevisionSchedule:
            schedules_filtered.append(schedule)
    return schedules_filtered


def get_schedules_by_built_in_category (doc, built_in_category):
    """
    Get all schedules in the current document by built-in category.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param built_in_category: Built-in category to filter schedules by.
    :type built_in_category: Autodesk.Revit.DB.BuiltInCategory

    :return: List of schedules in the current document filtered by built-in category.
    :rtype: list[Autodesk.Revit.DB.ViewSchedule]
    """
    
    schedules_filtered = []
    
    # get all schedules in the model
    schedules_in_model = get_schedules(doc)

    # get the category from the built-in category
    cat = Category.GetCategory(doc, built_in_category)

    # filter schedules by category
    for schedule in schedules_in_model:
        if schedule.Definition.CategoryId == cat.Id:
            schedules_filtered.append(schedule)
    
    return schedules_filtered


def get_all_multi_category_schedules(doc):
    """
    Get all multi-category schedules in the current document.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of multi-category schedules in the current document.
    :rtype: list[Autodesk.Revit.DB.ViewSchedule]
    """
    
    multi_category_schedules = []
    
    # get all schedules in the model
    schedules_in_model = get_schedules(doc)

    # filter for multi-category schedules (invalid category element id)
    for schedule in schedules_in_model:
        if schedule.Definition.CategoryId == ElementId.InvalidElementId:
            multi_category_schedules.append(schedule)
    
    return multi_category_schedules


def get_all_sheet_schedules(doc):
    """
    Get all schedules that are placed on sheets in the current document.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of schedules that are placed on sheets.
    :rtype: list[Autodesk.Revit.DB.ViewSchedule]
    """
    
    sheet_schedules = get_schedules_by_built_in_category (
        doc=doc, 
        built_in_category=BuiltInCategory.OST_Sheets
    )
    
    return sheet_schedules


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Model health report functions - groupes.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Model health report metrics can either be displayed in a family where each parameter is assigned to a metric 
and or data can be exported to text files which can be used to visualize key metrics over time.

"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr


import System

clr.AddReference("System.Core")
from System import Linq

clr.ImportExtensions(Linq)

from duHast.Revit.Common.groups import (
    get_detail_groups,
    get_model_groups,
    get_unplaced_detail_groups,
    get_unplaced_model_groups,
)

from duHast.Revit.ModelHealth.Reporting.Properties.constants import (
    FAILED_TO_RETRIEVE_VALUE,
)

# ---------------------------------------------  Groups  ---------------------------------------------


def get_number_of_detail_groups(doc):
    """
    Gets the number of detail group definitions the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: Number of detail group definitions in the model. On exception it will return -1
    :rtype: int
    """

    number = FAILED_TO_RETRIEVE_VALUE
    try:
        number = len(get_detail_groups(doc))
    except Exception as e:
        raise ValueError("Failed to get number of detail groups: {}".format(e))
    return number


def get_number_of_model_groups(doc):
    """
    Gets the number of model group definitions in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: Number of model group definitions in the model. On exception it will return -1
    :rtype: int
    """

    number = FAILED_TO_RETRIEVE_VALUE
    try:
        number = len(get_model_groups(doc))
    except Exception as e:
        raise ValueError("Failed to get number of model groups: {}".format(e))
    return number


def get_number_of_unplaced_detail_groups(doc):
    """
    Gets the number of unplaced detail group definitions in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: Number of unplaced detail group definitions in the model. On exception it will return -1
    :rtype: int
    """

    number = FAILED_TO_RETRIEVE_VALUE
    try:
        number = len(get_unplaced_detail_groups(doc))
    except Exception as e:
        raise ValueError("Failed to get number of unplaced detail groups: {}".format(e))
    return number


def get_number_of_unplaced_model_groups(doc):
    """
    Gets the number of unplaced model group definitions in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: Number of unplaced model group definitions in the model. On exception it will return -1
    :rtype: int
    """

    number = FAILED_TO_RETRIEVE_VALUE
    try:
        number = len(get_unplaced_model_groups(doc))
    except Exception as e:
        raise ValueError("Failed to get number of unplaced model groups: {}".format(e))
    return number



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit sub-category helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from Autodesk.Revit.DB import Transaction


from duHast.Utilities.Objects import result as res
from duHast.Revit.Categories.categories import (
    delete_main_sub_category,
    does_main_sub_category_exists,
    get_main_sub_categories,
)
from duHast.Revit.Categories.Utility.elements_by_category_utils import (
    move_elements_from_sub_category_to_sub_category,
)
from duHast.Revit.Categories.Utility.category_properties_set_utils import (
    set_category_properties,
)
from duHast.Revit.Categories.Utility.category_properties_get_utils import (
    get_category_properties,
)
from duHast.Revit.Common import transaction as rTran


def create_new_sub_category_to_family_category(doc, new_sub_category_name):
    """
    Creates a new subcategory to the family category and returns it.

    Note: if a subcategory with the name provided already exist it will be returned instead of trying to create another one with the same name.

    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document
    :param new_sub_category_name: The new subcategory name
    :type new_sub_category_name: str
    :return: The new subcategory. Exception "The name 'xys' is already in use" if subcategory with the same name is already in file.
    :rtype: A category. (or str if exception occurred)
    """

    return_value = res.Result()
    if doc.IsFamilyDocument:
        # check if subcategory already exists
        if does_main_sub_category_exists(doc, new_sub_category_name):
            # just return the already existing subcategory
            main_sub_categories = get_main_sub_categories(doc)
            return_value.update_sep(True, "Subcategory already in family.")
            return_value.result = main_sub_categories[new_sub_category_name]
        else:
            # create a new subcategory
            # get the family category
            current_fam_cat = doc.OwnerFamily.FamilyCategory
            parent_category = None
            # get parent category object from Revit internal settings
            for main_cat in doc.Settings.Categories:
                if main_cat.Name == current_fam_cat.Name:
                    parent_category = main_cat
                    break
            if new_sub_category_name != parent_category.Name:

                def action():
                    action_return_value = res.Result()
                    try:
                        new_sub_category = doc.Settings.Categories.NewSubcategory(
                            parent_category, new_sub_category_name
                        )
                        action_return_value.update_sep(
                            True,
                            "Created subcategory: {}".format(new_sub_category_name),
                        )
                        action_return_value.result = new_sub_category
                    except Exception as e:
                        action_return_value.update_sep(
                            False,
                            "Failed to create: {} with exception: {}".format(
                                new_sub_category_name, e
                            ),
                        )
                    return action_return_value

                transaction = Transaction(
                    doc, "Creating subcategory: {}".format(new_sub_category_name)
                )
                return_value = rTran.in_transaction(transaction, action)
            else:
                return_value.update_sep(
                    False,
                    "Cant create subcategory with the same name as the family category!",
                )
    else:
        return_value.update_sep(False, "This is not a family document!")
    return return_value


def create_new_category_from_saved_properties(
    doc, new_cat_name, saved_cat_props, ignore_missing_cut_style=False
):
    """
    Creates a new category and applies properties stored.

    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document
    :param new_cat_name: The new sub category name.
    :type new_cat_name: str
    :param saved_cat_props: Dictionary containing subcategory properties.
    :type saved_cat_props: list of dictionaries in format as per GetCategoryProperties(cat) method.
    :param ignore_missing_cut_style: If true will not flag an exception if applying styles fails on missing cut style.
    :type ignore_missing_cut_style: bool
    :return:
        Result class instance.
        - result.status. True if category was created or already existed in file, otherwise False.
        - result.message will contain the name of the category created.
        - result.result returns new category, if category already exists in file it will return that
        On exception:
        - result.status (bool) will be False.
        - result.message will contain generic exception message.
        - result.result will be empty
    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    result_new_sub_cat = create_new_sub_category_to_family_category(doc, new_cat_name)
    if result_new_sub_cat.status:
        new_sub_cat = result_new_sub_cat.result
        flag = set_category_properties(
            doc, new_sub_cat, saved_cat_props, ignore_missing_cut_style
        )
        if flag:
            return_value.update_sep(
                True, "Successfully created category: {}".format(new_cat_name)
            )
            return_value.result = new_sub_cat
        else:
            return_value.update_sep(
                False,
                "Failed to apply properties to new category: {}".format(new_cat_name),
            )
    else:
        return_value.update_sep(
            False, "Failed to create new subcategory: {}".format(new_cat_name)
        )
    return return_value


def create_new_category_and_transfer_properties(doc, new_cat_name, existing_cat_name):
    """
    Creates a new subcategory and transfer properties from existing subcategory.

    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document
    :param new_cat_name: The new sub category name.
    :type new_cat_name: str
    :param existing_cat_name: The existing subcategory name
    :type existing_cat_name: str
    :return:
        Result class instance.
        - result.status. True if category was created or already existed in file, otherwise False.
        - result.message will contain the name of the category created.
        - result.result returns new category, if category already exists in file it will return that
        On exception:
        - result.status (bool) will be False.
        - result.message will contain generic exception message.
        - result.result will be empty
    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    cats = get_main_sub_categories(doc)
    # check if existing category actually exists in family
    if existing_cat_name in cats:
        # check whether the new category already exists!
        if new_cat_name not in cats:
            copy_from_cat = cats[existing_cat_name]
            cat_props = get_category_properties(copy_from_cat, doc)
            result_new_sub_cat = create_new_category_from_saved_properties(
                doc, new_cat_name, cat_props
            )
            return_value.update(result_new_sub_cat)
        else:
            return_value.update_sep(
                True, "Category already in file: {}".format(new_cat_name)
            )
            return_value.result = cats[new_cat_name]
    else:
        return_value.update_sep(
            False,
            "Template category: {} does not exist in file!".format(existing_cat_name),
        )
    return return_value


def rename_sub_category(doc, old_sub_cat_name, new_sub_cat_name):
    """
    Renames a family custom subcategory.

    Note: Only subcategories directly belonging to the family category will be checked for a match.

    - The Revit API currently does not allow changing a subcategory name. This method instead:
    - Duplicates the old subcategory with the new name.
    - Moves all elements belonging to the old subcategory to the new subcategory.
    - Deletes the old subcategory.
    - If the new subcategory already exists in the file:
    - Moves all elements belonging to the old subcategory to the new subcategory.
    - Deletes the old subcategory.

    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document
    :param old_sub_cat_name: The subcategory name to be renamed.
    :type old_sub_cat_name: str
    :param new_sub_cat_name: The new subcategory name.
    :type new_sub_cat_name: str

    :return: Result class instance.
        - `result.status` (bool): True if the subcategory was renamed successfully, otherwise False.
        - `result.message` (str): Rename process messages.
        - `result.result` (list): Empty list.

    On exception:
        - `result.status` (bool): False.
        - `result.message` (str): Generic exception message.
        - `result.result` (list): Empty.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    # check whether ol;d subcategory exists in family file
    already_in_family_old = does_main_sub_category_exists(doc, old_sub_cat_name)
    if already_in_family_old:
        # check whether new subcategory already in family
        already_in_family = does_main_sub_category_exists(doc, new_sub_cat_name)
        if already_in_family:
            # just move elements from old sub category to new one
            return_value.append_message(
                "Subcategory: {} already in family.".format(new_sub_cat_name)
            )
        else:
            # duplicate old sub category
            create_new_status = create_new_category_and_transfer_properties(
                doc, new_sub_cat_name, old_sub_cat_name
            )
            return_value.update(create_new_status)
        # check if we have a subcategory to move elements to
        if return_value.status:
            # move elements
            move_status = move_elements_from_sub_category_to_sub_category(
                doc, old_sub_cat_name, new_sub_cat_name
            )
            return_value.update(move_status)
            if move_status.status:
                deleted_old_sub_category = delete_main_sub_category(
                    doc, old_sub_cat_name
                )
                if deleted_old_sub_category:
                    return_value.update_sep(
                        True,
                        "Subcategory: {} deleted successfully.".format(
                            old_sub_cat_name
                        ),
                    )
                else:
                    return_value.update_sep(
                        True,
                        "Subcategory: {} failed to delete subcategory...Exiting".format(
                            old_sub_cat_name
                        ),
                    )
            else:
                return_value.update_sep(
                    False,
                    "Subcategory: {} failed to move elements to new subcategory. Exiting...".format(
                        new_sub_cat_name
                    ),
                )
        else:
            return_value.update_sep(
                False,
                "Subcategory: {} failed to create in family. Exiting...".format(
                    new_sub_cat_name
                ),
            )
    else:
        return_value.update_sep(
            False,
            "Base subcategory: {} does not exist in family. Exiting...".format(
                old_sub_cat_name
            ),
        )
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a function to swap out one shard parameter for another.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

The swap is done by:
- changing current shared parameter to a family parameter (dummy)
- deleting the old shared parameter definition
- swapping the family parameter (dummy) to the new shared parameter.

Note: storage types of old and new shared parameter need to be identical.

Parameter change directives are read from a .csv file:

- header row: yes
- column 1: current shared parameter name
- column 2: new shared parameter name
- column 3: fully qualified file path of shared parameter file
- column 4: Is parameter instance (True / False)
- column 5: parameter grouping name ( refer to module: RevitParameterGrouping)

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import System

# import common library modules
# from duHast.APISamples.Common import RevitCommonAPI as com
from duHast.Utilities.Objects import result as res
from duHast.Utilities import files_csv as fileCSV
from duHast.Utilities.files_io import file_exist
from duHast.Revit.SharedParameters import shared_parameter_add as rSharedPAdd
from duHast.Revit.SharedParameters import shared_parameters_tuple as rSharedT
from duHast.Revit.Common import parameter_grouping as rPG
from duHast.Revit.SharedParameters import shared_parameters as rSharedPara
from duHast.Revit.SharedParameters import shared_parameters_delete as rSharedParaDelete
from duHast.Revit.SharedParameters import (
    shared_parameter_type_change as rSharedTypeChange,
)

from collections import namedtuple


"""
Tuple containing settings data on how to swap a shared parameter retrieved from a file.
"""

PARAMETER_SETTINGS_DATA = namedtuple(
    "parameterSettingsData", "oldParameterName newParameterData sharedParameterPath"
)


def _load_shared_parameter_data_from_file(file_path):
    """
    Reads parameter data from file and convertes it into a dictionary of named tuples (PARAMETER_SETTINGS_DATA)

    :param file_path: Fully qualified file path to shared parameter change directive file.
    :type file_path: str

    :return: A dictionary in format; key: current parameter name, value: named tuple of type parameterSettingsData
    :rtype: {str:named tuple}
    """

    parameter_mapper = {}
    
    # attempt to read file
    file_data_result = fileCSV.read_csv_file(file_path)
    if file_data_result.status == False:
        raise Exception(file_data_result.message)
    
    file_data = file_data_result.result
    
    for i in range(1, len(file_data)):
        row = file_data[i]
        if len(row) == 5:
            flag = False
            if row[3].upper() == "TRUE":
                flag = True
            t = rSharedT.PARAMETER_DATA(
                row[1], flag, rPG.PRAMETER_GROPUING_TO_BUILD_IN_PARAMETER_GROUPS[row[4]]
            )
            parameter_mapper[row[0]] = PARAMETER_SETTINGS_DATA(row[0], t, row[2])
    return parameter_mapper


def swap_shared_parameters(doc, change_directive_file_path):
    """
    Swaps out a shared parameter for another. (refer to module header for details)

    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document
    :param change_directive_file_path: Fully qualified file path to shared parameter change directive.
    :type change_directive_file_path: str

    :return:
        Result class instance.

        - False if an exception occurred, otherwise True.
        - result.message will contain the names of the changed shared parameter(s).
        - result status will contain lists of new shared parameters

        On exception (handled by optimizer itself!):

        - result.status (bool) will be False.
        - result.message will contain exception message.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    _parameter_prefix_ = "_dummy_"

    # check if file exists
    if file_exist(change_directive_file_path) == False:
        return_value.update_sep(
            False,
            "Change directive file does not exist: {}".format(
                change_directive_file_path
            ),
        )
        return return_value

    # load change directive
    parameter_directives = _load_shared_parameter_data_from_file(
        change_directive_file_path
    )

    # check if there are any directives
    if len(parameter_directives) == 0:
        return_value.update_sep(False, "No parameter directives in file.")
        return return_value

    # loop over directive and
    for p_directive in parameter_directives:
        # load shared para file
        shared_para_def_file = rSharedPAdd.load_shared_parameter_file(
            doc, parameter_directives[p_directive].sharedParameterPath
        )
        return_value.append_message(
            "Read shared parameter file: {}".format(
                parameter_directives[p_directive].sharedParameterPath
            )
        )

        if shared_para_def_file != None:
            #   - swap shared parameter to family parameter
            status_change_to_fam_para = (
                rSharedTypeChange.change_shared_parameter_to_family_parameter(
                    doc, p_directive, _parameter_prefix_
                )
            )
            return_value.update(status_change_to_fam_para)
            if status_change_to_fam_para.status:
                #   - delete all shared parameter definition
                status_delete_old_shared_para_def = (
                    rSharedParaDelete.delete_shared_parameter_by_name(doc, p_directive)
                )
                return_value.update(status_delete_old_shared_para_def)
                if status_delete_old_shared_para_def.status:
                    # get shared parameter definition
                    s_para_def = rSharedPara.get_shared_parameter_definition(
                        parameter_directives[p_directive].newParameterData.name,
                        shared_para_def_file,
                    )
                    #   - add new shared parameter
                    if s_para_def != None:
                        return_value.append_message(
                            "Retrieved shared parameter definition for: {}".format(
                                parameter_directives[p_directive].newParameterData.name
                            )
                        )
                        #   - swap family parameter to shared parameter
                        status_swap_fam_to_shared_p = rSharedTypeChange.change_family_parameter_to_shared_parameter(
                            doc,
                            _parameter_prefix_ + p_directive,  # add prefix
                            parameter_directives[p_directive].newParameterData,
                            s_para_def,
                        )
                        return_value.update(status_swap_fam_to_shared_p)
                    else:
                        return_value.update_sep(
                            False,
                            "Failed to get shared parameter definition from file.",
                        )
                else:
                    return_value.update(status_delete_old_shared_para_def)
            else:
                return_value.update(status_change_to_fam_para)
        else:
            return_value.update_sep(
                False,
                "Failed to load shared parameter def file from: {}".format(
                    parameter_directives[p_directive].sharedParameterPath
                ),
            )

    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Geometry data storage class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import json
from duHast.Data.Objects.Collectors import data_base
from duHast.Data.Objects.Collectors.Properties.data_property_names import DataPropertyNames
from duHast.Geometry.point_3 import Point3
from duHast.Geometry.matrix import Matrix

class DataGeometryBase(data_base.DataBase):
    def __init__(self, data_type, j=None, **kwargs):
        """
        Class constructor

        :param j:  json formatted dictionary of this class, defaults to {}
        :type j: dict, optional
        """
        
        # store data type  in base class
        super(DataGeometryBase, self).__init__(data_type=data_type, j=j, **kwargs)

        # set default values
        # translation as per shared coordinates in revit file
        self.translation_coord = Point3(0.0, 0.0, 0.0)
        # rotation as per shared coordinates in revit file ( default )
        self.rotation_coord = Matrix(rows=3, cols=3)

        json_var=None
        # check if any data was past in with constructor!
        if j is not None:
            # check type of data that came in:
            if isinstance(j, str):
                # a string
                json_var = json.loads(j)
            elif isinstance(j, dict):
                # make a copy
                json_var=j.copy()
            else:
                raise TypeError(
                    "Argument j supplied must be of type string or type dictionary. Got {} instead.".format(
                        type(j)
                    )
                )

            # attempt to populate from json
            try:
                translation_coord = json_var.get(
                    DataPropertyNames.TRANSLATION_COORDINATES, None
                )
                # check if we got None back...if so use what is the default
                # since a point can be initialized with None
                if translation_coord is not None:
                    self.translation_coord = Point3(j=translation_coord)

                rotation_coord = json_var.get(
                    DataPropertyNames.ROTATION_COORDINATES, None
                )
                # check if we got None back...if so use what is the default
                # since a matrix ini from an empty dictionary got 0 x 0 size, meanwhile our default is 3 x 3
                if rotation_coord is not None:
                    self.rotation_coord = Matrix(j=rotation_coord)

            except Exception as e:
                raise type(e)(
                    "Node {} failed to initialise with: {}".format(self.data_type, e)
                )

    def __eq__(self, other):
        if not isinstance(other, DataGeometryBase):
            return NotImplemented
        return (
            self.translation_coord == other.translation_coord
            and self.rotation_coord == other.rotation_coord
        )

    def __ne__(self, other):
        return not self.__eq__(other)
    
    def __hash__(self):
        return hash((self.translation_coord, self.rotation_coord))



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family data class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

- contains:

    - an entire family nesting tree structure represented as other  FamilyDataFamily instances
    - all FamilyDataContainer instances belonging to this family only ( identified by unique combination on family name nesting path and family category nesting path )

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects import base
from duHast.Revit.Family.Data.Objects.family_data_container import FamilyDataContainer
from duHast.Revit.Family.Data.Objects.family_base_data_processor_defaults import (
    NESTING_SEPARATOR,
)


class FamilyDataFamily(base.Base):
    def __init__(
        self,
        family_name=None,
        family_category=None,
        family_file_path=None,
        family_nesting_path=None,
        family_category_nesting_path=None,
        is_root_family=False,
        nested_families=None,
    ):
        """
        Family data class.

        - contains an entire family nesting tree structure represented as family data container instances

        :param family_name: The name of the family.
        :type family_name: str
        :param family_category: The Revit category of the family.
        :type family_category: str
        :param family_file_path: The file path of the family.
        :type family_file_path: str

        """

        # forwards all unused arguments
        # ini super class to allow multi inheritance in children!
        super(FamilyDataFamily, self).__init__()

        # set some base properties to their default values
        if isinstance(family_name, str) or family_name == None:
            self.family_name = family_name
        else:
            raise TypeError(
                "family_name must either be a string or None. Got: {}".format(
                    type(family_name)
                )
            )

        if isinstance(family_category, str) or family_category == None:
            self.family_category = family_category
        else:
            raise TypeError(
                "family_category must either be a string or None. Got: {}".format(
                    type(family_category)
                )
            )

        if isinstance(family_file_path, str) or family_file_path == None:
            self.family_file_path = family_file_path
        else:
            raise TypeError(
                "family_file_path must either be a string or None. Got {}".format(
                    type(family_file_path)
                )
            )

        if isinstance(family_nesting_path, str) or family_nesting_path == None:
            self.family_nesting_path = family_nesting_path
        else:
            raise TypeError(
                "family_nesting_path must either be a string or None. Got {}".format(
                    type(family_nesting_path)
                )
            )

        if (
            isinstance(family_category_nesting_path, str)
            or family_category_nesting_path == None
        ):
            self.family_category_nesting_path = family_category_nesting_path
        else:
            raise TypeError(
                "family_category_nesting_path must either be a string or None. Got {}".format(
                    type(family_category_nesting_path)
                )
            )

        if isinstance(is_root_family, bool):
            self.is_root_family = is_root_family
        else:
            raise TypeError(
                "is_root_family must be a boolean. Got {}".format(type(is_root_family))
            )

        # nested family tree
        self.nested_families_unsorted = []
        if nested_families is not None:
            for nested in nested_families:
                if not isinstance(nested, FamilyDataFamily):
                    raise ValueError(
                        "nested_families must be a list of FamilyDataFamily instances. Got: {}".format(
                            type(nested)
                        )
                    )
                self.nested_families_unsorted.append(nested)

        # default value for data containers
        self.data_containers_unsorted = []

        # default value for nesting by name path
        self.nesting_by_name_path = {}

        # default value for nesting by level
        self.nesting_by_level = {}

        # flag indicating whether nodes have been processed
        self.is_processed = False

    def __repr__(self):
        return "{}<{}>".format(self.family_name, self.family_category)

    def __eq__(self, other):
        """
        Compares family_name, family_category, family_nesting_path and family_category_nesting_path properties only

        :param other: Another FamilyDataFamily instance
        :type other: :class:`.FamilyDataFamily`
        :return: True if family_name and family_category are equal otherwise false
        :rtype: bool
        """

        return isinstance(other, FamilyDataFamily) and (
            self.family_name,
            self.family_category,
            self.family_nesting_path,
            self.family_category_nesting_path,
        ) == (
            other.family_name,
            other.family_category,
            other.family_nesting_path,
            other.family_category_nesting_path,
        )

    def compare_name_and_category(self, other):
        """
        Compares family_name and family_category properties only

        :param other: Another FamilyDataFamily instance
        :type other: :class:`.FamilyDataFamily`
        :return: True if family_name and family_category are equal otherwise false
        :rtype: bool
        """
        return isinstance(other, FamilyDataFamily) and (
            self.family_name,
            self.family_category,
        ) == (
            other.family_name,
            other.family_category,
        )


    # python 2.7 needs custom implementation of not equal
    def __ne__(self, other):
        return not self.__eq__(other=other)


    def __hash__(self):
        """
        Hash the family data.
        
        Hash only considers the family_name, family_category, family_nesting_path and family_category_nesting_path properties.
        
        :return: Hash of the family data
        :rtype: int
        """
        
        return hash(
            (self.family_name,
            self.family_category,
            self.family_nesting_path,
            self.family_category_nesting_path,
            )
        )
    
    def _build_nesting_by_name(self):
        """
        Build the nesting name for the family data.

        """
        self.nesting_by_name_path = {}

        for family_instance_data in self.nested_families_unsorted:
            self.nesting_by_name_path[
                family_instance_data.family_nesting_path
            ] = family_instance_data

    def _build_nesting_by_level(self):
        """
        Build the nesting by level for the family data.

        """

        self.nesting_by_level = {}
        # start at 1 because nesting level 0 is this family (root)
        # ( 1 based rather then 0 based )
        for family_instance_data in self.nested_families_unsorted:
            if family_instance_data.family_nesting_path == None:
                raise ValueError(
                    "Family nesting path is None. Cannot build nesting by level."
                )
            elif family_instance_data.family_nesting_path == "":
                raise ValueError(
                    "Family nesting path is empty. Cannot build nesting by level."
                )
            else:
                nesting_chunks = family_instance_data.family_nesting_path.split(
                    NESTING_SEPARATOR
                )

                if len(nesting_chunks) - 1 in self.nesting_by_level:
                    self.nesting_by_level[len(nesting_chunks) - 1].append(
                        family_instance_data
                    )
                else:
                    self.nesting_by_level[len(nesting_chunks) - 1] = [
                        family_instance_data
                    ]

    def get_longest_unique_nesting_path(self):
        """
        Get the longest unique family name nesting path(s) and family category nesting path(s) of the family data.

        :return: The longest unique nesting path as a list of tuples, or None if family is not processed.
        :rtype: list[(family_name_nesting, family_category_nesting)] or None
        """

        unique_nesting_paths = []
        if self.is_processed == False:
            return

        # loop over nesting by level property and get the longest unique family name nesting path (multiple)
        # and category nesting path as a tuple where 0 index is the name path, and 1 index the category nesting path
        # nesting by level has the nesting depth as key and the data containers as value
        # start from the highest nesting level and check whether the next level down overlaps with the current level
        # all path in the highest (deepest) level of nesting are unique
        # if the next level down overlaps with the current level, the path in the next level down is not unique
        # if the next level down does not overlap with the current level, the path in the next level down is unique
        # as soon as there is a nested family, the root path of a family (no nesting) is not unique anymore.

        # get the highest nesting level
        if len(self.nesting_by_level) == 0:
            # if no nesting levels found return the root family only
            unique_nesting_paths.append(
                (self.family_nesting_path, self.family_category_nesting_path)
            )
        else:
            highest_nesting_level = max(self.nesting_by_level.keys())

            # get the data containers at the highest nesting level
            family_data_instances = self.nesting_by_level[highest_nesting_level]
            if family_data_instances == None:
                raise ValueError("No data containers found at highest nesting level.")

            # add their nesting path to the unique nesting paths
            for family_data_instance in family_data_instances:
                # add tuple made up off family name nesting path at 0 and family category nesting path at 1
                unique_nesting_paths.append(
                    (
                        family_data_instance.family_nesting_path,
                        family_data_instance.family_category_nesting_path,
                    )
                )

            # loop over the nesting levels from the highest to the lowest
            for nesting_level in range(highest_nesting_level - 1, 0, -1):
                # get the data containers at the current nesting level
                family_data_instances = self.nesting_by_level[nesting_level]
                if family_data_instances == None:
                    raise ValueError(
                        "No data containers found at nesting level: {}".format(
                            nesting_level
                        )
                    )
                # loop over the data containers at the current nesting level
                for family_data_instance in family_data_instances:
                    # get the nesting path of the current data container
                    nesting_path_family_name_current_level = (
                        family_data_instance.family_nesting_path
                    )
                    nesting_path_category_current_level = (
                        family_data_instance.family_category_nesting_path
                    )
                    # check for match of both path
                    for unique_nesting_path in unique_nesting_paths:
                        found_match = False
                        # check if the current nesting path overlaps with the unique nesting path
                        if (
                            nesting_path_family_name_current_level
                            in unique_nesting_path[0]
                            and nesting_path_category_current_level
                            in unique_nesting_path[1]
                        ):
                            # if it overlaps, remove it from the unique nesting paths
                            found_match = True
                            break
                    if not found_match:
                        unique_nesting_paths.append(
                            (
                                nesting_path_family_name_current_level,
                                nesting_path_category_current_level,
                            )
                        )
        # return the unique nesting paths
        return unique_nesting_paths

    def process(self):
        """
        Process the family data.

        - build the nesting by name path
        - build the nesting by level

        """

        # check if processed
        if self.is_processed:
            return

        # build internal data structures
        self._build_nesting_by_name()
        self._build_nesting_by_level()

        # set processed flag
        self.is_processed = True

    def add_data_container(self, data_container):
        """
        Add a data container to the family data.

        """

        if not isinstance(data_container, FamilyDataContainer):
            raise TypeError(
                "data_container must be an instance of FamilyDataContainer. Got: {}".format(
                    type(data_container)
                )
            )

        self.data_containers_unsorted.append(data_container)

        # set flag indicating that the family data has changed and needs to be processed again
        self.is_processed = False

    def add_nested_family_instance(self, nested_family_instance):
        """
        Add a nested family instance.

        :param nested_family_instance: The family instance to be added
        :type nested_family_instance: :class:`.FamilyDataFamily`
        :raises TypeError: If nested_family_instance is not an instance of FamilyDataFamily a type error will be raised.
        """

        # check correct type
        if not isinstance(nested_family_instance, FamilyDataFamily):
            raise TypeError(
                "nested_family_instance must be an instance of FamilyDataFamily. Got: {}".format(
                    type(nested_family_instance)
                )
            )

        # check if nested family instance already exists...if so raise an exception
        if nested_family_instance in self.nested_families_unsorted:
            raise ValueError(
                "family instance {} {}  already exists in family".format(
                    nested_family_instance.family_name,
                    nested_family_instance.family_category,
                )
            )

        # add the instance
        self.nested_families_unsorted.append(nested_family_instance)

        # set flag indicating that the family data has changed and needs to be processed again
        self.is_processed = False

    def get_all_storage_data_as_strings(self):
        """
        Returns the data storage within each storage container within this family as well as within the nested families as a a list of string.

        :return: A dictionary where key is the data storage type, and value is a nested list of lists containing the data storage string value
        :rtype: {key:[[str]]}
        """

        # set up the return value
        data_storage_as_string = {}

        # loop over each container belonging directly to this family and get its dictionary representing the storage type and its values as a list o string
        for container in self.data_containers_unsorted:
            container_data_storage_dic = container.get_data_string_list()
            for key, value in container_data_storage_dic.items():
                # check if that storage key is already in the return dictionary
                if key in data_storage_as_string:
                    # if extend its value list by the new containers values list
                    data_storage_as_string[key].extend(value)
                else:
                    # set up a new key and value list
                    data_storage_as_string[key] = value

        # loop over nested families and get their storage values too
        for nested_family in self.nested_families_unsorted:
            family_data_storage_dic = nested_family.get_all_storage_data_as_strings()
            for key, value in family_data_storage_dic.items():
                # check if that storage key is already in the return dictionary
                if key in data_storage_as_string:
                    # if extend its value list by the new containers values list
                    data_storage_as_string[key].extend(value)
                else:
                    # set up a new key and value list
                    data_storage_as_string[key] = value

        return data_storage_as_string

    def get_all_storage_headers_as_strings(self):
        """
        Returns all property names used in storage instances in containers within this family and any nested families.

        :return: A dictionary where key is the data storage type, and value is a list containing the data storage properties as string values
        :rtype: {key:[[str]]}
        """

        # set up the return value
        data_storage_as_string = {}
        # loop over each container belonging directly to this family and get its dictionary representing the storage type and its header values as a list o string
        for container in self.data_containers_unsorted:
            container_data_storage_dic = container.get_data_headers_list()
            for key, value in container_data_storage_dic.items():
                # check if that storage key is already in the return dictionary
                if key not in data_storage_as_string:
                    # if not add its value list by the new containers values list
                    data_storage_as_string[key] = value

        # loop over nested families and get their storage header values too
        for nested_family in self.nested_families_unsorted:
            family_data_storage_dic = nested_family.get_all_storage_headers_as_strings()
            for key, value in family_data_storage_dic.items():
                # check if that storage key is already in the return dictionary
                if key not in data_storage_as_string:
                    # if extend its value list by the new containers values list
                    data_storage_as_string[key] = value
        return data_storage_as_string

    def has_circular_nesting(self):
        """
        Check if the family data has circular nesting.

        Circular nesting is defined as a situation where a nesting path property of a container contains the same family more than once.

        returns a list of tuples in format:

            - 0 index the nesting level at which the circular nesting occurs
            - 1 index a string in format: family name :: family category
            - 2 the entire family name nesting path this circular reference appeared on

        """

        # check if processed and if not process the family data
        if not self.is_processed:
            self.process()

        # get the longest unique nesting path ( can be multiple )
        longest_unique_nesting_path = self.get_longest_unique_nesting_path()

        # keep record of duplicated family names
        duplicated_family_names = []
        # loop over these path and check for multiple occurrence of a family name
        for nesting_path in longest_unique_nesting_path:

            # split path at nesting separator ( should have the same length )
            nesting_chunks_family_names = nesting_path[0].split(NESTING_SEPARATOR)
            nesting_chunks_category = nesting_path[1].split(NESTING_SEPARATOR)

            # double check path length are equal!
            if len(nesting_chunks_family_names) != len(nesting_chunks_category):
                raise ValueError(
                    "Length of nesting name path: {} is not the same as length of nesting category path: {}".format(
                        len(nesting_chunks_family_names), len(nesting_chunks_category)
                    )
                )

            # check if a chunk is already in the list, indicating a circular nesting
            node_names = []
            for i in range(len(nesting_chunks_family_names)):

                # build a test value made from family name and category name
                test_value = (
                    nesting_chunks_family_names[i]
                    + NESTING_SEPARATOR
                    + nesting_chunks_category[i]
                )

                # check if test value is already in list...indicating a circular nesting reference
                if test_value in node_names:
                    # add family name and category to duplicated list including the nesting level
                    duplicated_family_names.append((i, test_value, nesting_path[0]))
                else:
                    # add test value to nodes list
                    node_names.append(test_value)

        return duplicated_family_names
